"""The DatasetDescriptor class."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Generator
from typing import List

from glotaran.model.attribute import model_attribute
from glotaran.parameter import Parameter

if TYPE_CHECKING:
    from glotaran.model.megacomplex import Megacomplex


@model_attribute(
    properties={
        "megacomplex": List[str],
        "megacomplex_scale": {"type": List[Parameter], "default": None, "allow_none": True},
        "scale": {"type": Parameter, "default": None, "allow_none": True},
    }
)
class DatasetDescriptor:
    """A `DatasetDescriptor` describes a dataset in terms of a glotaran model.
    It contains references to model items which describe the physical model for
    a given dataset.

    A general dataset describtor assigns one or more megacomplexes and a scale
    parameter.
    """

    def iterate_megacomplexes(self) -> Generator[tuple[Parameter | int, Megacomplex | str]]:
        for i, megacomplex in enumerate(self.megacomplex):
            scale = self.megacomplex_scale[i] if self.megacomplex_scale is not None else None
            yield scale, megacomplex



import numpy as np

from glotaran.analysis.util import calculate_matrix
from glotaran.builtin.models.kinetic_image import KineticImageModel
from glotaran.parameter import ParameterGroup


def test_baseline():
    model = KineticImageModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"type": "kinetic-decay", "k_matrix": ["k1"]},
                "mc2": {"type": "kinetic-baseline"},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1", "mc2"],
                },
            },
        }
    )

    parameter = ParameterGroup.from_list(
        [
            101e-4,
            [1, {"vary": False, "non-negative": False}],
            [42, {"vary": False, "non-negative": False}],
        ]
    )

    time = np.asarray(np.arange(0, 50, 1.5))
    dataset = model.dataset["dataset1"].fill(model, parameter)
    compartments, matrix = calculate_matrix(model, dataset, {}, {"time": time})

    assert len(compartments) == 2
    assert compartments[1] == "dataset1_baseline"

    assert matrix.shape == (time.size, 2)
    assert np.all(matrix[:, 1] == 1)



from __future__ import annotations

from typing import Any

import numpy as np

from glotaran.builtin.models.kinetic_image.initial_concentration import InitialConcentration
from glotaran.model import model_attribute
from glotaran.parameter import Parameter

class KMatrix:
    @classmethod
    def empty(cls: Any, label: str, compartments: list[str]) -> KMatrix: ...
    def involved_compartments(self) -> list[str]: ...
    def combine(self, k_matrix: KMatrix) -> KMatrix: ...
    def matrix_as_markdown(
        self, compartments: list[str] = ..., fill_parameters: bool = ...
    ) -> str: ...
    def a_matrix_as_markdown(self, initial_concentration: InitialConcentration) -> str: ...
    def reduced(self, compartments: list[str]) -> np.ndarray: ...
    def full(self, compartments: list[str]) -> np.ndarray: ...
    def eigen(self, compartments: list[str]) -> tuple[np.ndarray, np.ndarray]: ...
    def rates(self, initial_concentration: InitialConcentration) -> np.ndarray: ...
    def a_matrix(self, initial_concentration: InitialConcentration) -> np.ndarray: ...
    def a_matrix_non_unibranch(
        self, initial_concentration: InitialConcentration
    ) -> np.ndarray: ...
    def a_matrix_unibranch(self, initial_concentration: InitialConcentration) -> np.array: ...
    def is_unibranched(self, initial_concentration: InitialConcentration) -> bool: ...
    @property
    def matrix(self) -> dict[tuple[str, str], Parameter]: ...



from __future__ import annotations

import numpy as np

from glotaran.model import DatasetDescriptor
from glotaran.model import model_attribute


@model_attribute(has_type=True)
class Megacomplex:
    def calculate_matrix(
        self,
        model,
        dataset_descriptor: DatasetDescriptor,
        indices: dict[str, int],
        axis: dict[str, np.ndarray],
        **kwargs,
    ):
        raise NotImplementedError



"""This package contains the kinetic megacomplex item."""
from __future__ import annotations

from typing import List

import numba as nb
import numpy as np

from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.model import DatasetDescriptor
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import model_attribute


@model_attribute(
    properties={
        "k_matrix": List[str],
    }
)
class KineticDecayMegacomplex(Megacomplex):
    """A Megacomplex with one or more K-Matrices."""

    def has_k_matrix(self) -> bool:
        return len(self.k_matrix) != 0

    def full_k_matrix(self, model=None):
        full_k_matrix = None
        for k_matrix in self.k_matrix:
            if model:
                k_matrix = model.k_matrix[k_matrix]
            if full_k_matrix is None:
                full_k_matrix = k_matrix
            # If multiple k matrices are present, we combine them
            else:
                full_k_matrix = full_k_matrix.combine(k_matrix)
        return full_k_matrix

    @property
    def involved_compartments(self):
        return self.full_k_matrix().involved_compartments() if self.full_k_matrix() else []

    def calculate_matrix(
        self,
        model,
        dataset_descriptor: DatasetDescriptor,
        indices: dict[str, int],
        axis: dict[str, np.ndarray],
        **kwargs,
    ):
        if dataset_descriptor.initial_concentration is None:
            raise ModelError(
                f'No initial concentration specified in dataset "{dataset_descriptor.label}"'
            )
        initial_concentration = dataset_descriptor.initial_concentration.normalized()

        k_matrix = self.full_k_matrix()

        # we might have more compartments in the model then in the k matrix
        compartments = [
            comp
            for comp in initial_concentration.compartments
            if comp in k_matrix.involved_compartments()
        ]

        # the rates are the eigenvalues of the k matrix
        rates = k_matrix.rates(initial_concentration)

        global_index = indices.get(model.global_dimension, None)
        global_axis = axis.get(model.global_dimension, None)
        model_axis = axis[model.model_dimension]

        # init the matrix
        size = (model_axis.size, rates.size)
        matrix = np.zeros(size, dtype=np.float64)

        kinetic_image_matrix_implementation(
            matrix, rates, global_index, global_axis, model_axis, dataset_descriptor
        )

        if not np.all(np.isfinite(matrix)):
            raise ValueError(
                f"Non-finite concentrations for K-Matrix '{k_matrix.label}':\n"
                f"{k_matrix.matrix_as_markdown(fill_parameters=True)}"
            )

        # apply A matrix
        matrix = matrix @ k_matrix.a_matrix(initial_concentration)

        # done
        return (compartments, matrix)


def kinetic_image_matrix_implementation(
    matrix: np.ndarray,
    rates: np.ndarray,
    global_index: int,
    global_axis: np.ndarray,
    model_axis: np.ndarray,
    dataset_descriptor: DatasetDescriptor,
):
    if isinstance(dataset_descriptor.irf, IrfMultiGaussian):

        (
            centers,
            widths,
            irf_scales,
            shift,
            backsweep,
            backsweep_period,
        ) = dataset_descriptor.irf.parameter(global_index, global_axis)

        for center, width, irf_scale in zip(centers, widths, irf_scales):
            calculate_kinetic_matrix_gaussian_irf(
                matrix,
                rates,
                model_axis,
                center - shift,
                width,
                irf_scale,
                backsweep,
                backsweep_period,
            )
        if dataset_descriptor.irf.normalize:
            matrix /= np.sum(irf_scale)

    else:
        calculate_kinetic_matrix_no_irf(matrix, rates, model_axis)


@nb.jit(nopython=True, parallel=True)
def calculate_kinetic_matrix_no_irf(matrix, rates, times):
    for n_r in nb.prange(rates.size):
        r_n = rates[n_r]
        for n_t in range(times.size):
            t_n = times[n_t]
            matrix[n_t, n_r] += np.exp(r_n * t_n)


sqrt2 = np.sqrt(2)


@nb.jit(nopython=True, parallel=True)
def calculate_kinetic_matrix_gaussian_irf(
    matrix, rates, times, center, width, scale, backsweep, backsweep_period
):
    """Calculates a kinetic matrix with a gaussian irf."""
    for n_r in nb.prange(rates.size):
        r_n = -rates[n_r]
        backsweep_valid = abs(r_n) * backsweep_period > 0.001
        alpha = (r_n * width) / sqrt2
        for n_t in nb.prange(times.size):
            t_n = times[n_t]
            beta = (t_n - center) / (width * sqrt2)
            thresh = beta - alpha
            if thresh < -1:
                matrix[n_t, n_r] += scale * 0.5 * erfcx(-thresh) * np.exp(-beta * beta)
            else:
                matrix[n_t, n_r] += (
                    scale * 0.5 * (1 + erf(thresh)) * np.exp(alpha * (alpha - 2 * beta))
                )
            if backsweep and backsweep_valid:
                x1 = np.exp(-r_n * (t_n - center + backsweep_period))
                x2 = np.exp(-r_n * ((backsweep_period / 2) - (t_n - center)))
                x3 = np.exp(-r_n * backsweep_period)
                matrix[n_t, n_r] += scale * (x1 + x2) / (1 - x3)


import ctypes  # noqa: E402

# This is a work around to use scipy.special function with numba
from numba.extending import get_cython_function_address  # noqa: E402

_dble = ctypes.c_double

functype = ctypes.CFUNCTYPE(_dble, _dble)

erf_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erf")
erfcx_addr = get_cython_function_address("scipy.special.cython_special", "__pyx_fuse_1erfcx")

erf = functype(erf_addr)
erfcx = functype(erfcx_addr)



from __future__ import annotations

import numpy as np
import xarray as xr

from glotaran.analysis.problem import Problem
from glotaran.analysis.problem import UngroupedProblemDescriptor
from glotaran.analysis.util import calculate_matrix
from glotaran.analysis.util import reduce_matrix


class UngroupedProblem(Problem):
    """Represents a problem where the data is not grouped."""

    def init_bag(self):
        """Initializes an ungrouped problem bag."""
        self._bag = {}
        for label in self._scheme.model.dataset:
            dataset = self._data[label]
            data = dataset.data
            weight = dataset.weight if "weight" in dataset else None
            if weight is not None:
                data = data * weight
                dataset["weighted_data"] = data
            self._bag[label] = UngroupedProblemDescriptor(
                self._scheme.model.dataset[label],
                data,
                dataset.coords[self._model_dimension].values,
                dataset.coords[self._global_dimension].values,
                weight,
            )

    def calculate_index_dependent_matrices(
        self,
    ) -> tuple[
        dict[str, list[list[str]]],
        dict[str, list[np.ndarray]],
        dict[str, list[str]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the index dependent model matrices."""

        self._clp_labels = {}
        self._matrices = {}
        self._reduced_clp_labels = {}
        self._reduced_matrices = {}

        for label, problem in self.bag.items():
            self._clp_labels[label] = []
            self._matrices[label] = []
            self._reduced_clp_labels[label] = []
            self._reduced_matrices[label] = []
            descriptor = self._filled_dataset_descriptors[label]

            for i, index in enumerate(problem.global_axis):
                result = calculate_matrix(
                    self._model,
                    descriptor,
                    {self._global_dimension: i},
                    {
                        self._model_dimension: problem.model_axis,
                        self._global_dimension: problem.global_axis,
                    },
                )

                self._clp_labels[label].append(result.clp_label)
                self._matrices[label].append(result.matrix)
                reduced_labels_and_matrix = reduce_matrix(
                    self._model, label, self._parameters, result, index
                )
                self._reduced_clp_labels[label].append(reduced_labels_and_matrix.clp_label)
                self._reduced_matrices[label].append(reduced_labels_and_matrix.matrix)

        return self._clp_labels, self._matrices, self._reduced_clp_labels, self._reduced_matrices

    def calculate_index_independent_matrices(
        self,
    ) -> tuple[
        dict[str, list[str]],
        dict[str, np.ndarray],
        dict[str, list[str]],
        dict[str, np.ndarray],
    ]:
        """Calculates the index independent model matrices."""

        self._clp_labels = {}
        self._matrices = {}
        self._reduced_clp_labels = {}
        self._reduced_matrices = {}

        for label, descriptor in self._filled_dataset_descriptors.items():
            model_axis = self._data[label].coords[self._model_dimension].values
            global_axis = self._data[label].coords[self._global_dimension].values
            result = calculate_matrix(
                self._model,
                descriptor,
                {},
                {
                    self._model_dimension: model_axis,
                    self._global_dimension: global_axis,
                },
            )

            self._clp_labels[label] = result.clp_label
            self._matrices[label] = result.matrix
            reduced_result = reduce_matrix(self._model, label, self._parameters, result, None)
            self._reduced_clp_labels[label] = reduced_result.clp_label
            self._reduced_matrices[label] = reduced_result.matrix

        return self._clp_labels, self._matrices, self._reduced_clp_labels, self._reduced_matrices

    def calculate_index_dependent_residual(
        self,
    ) -> tuple[
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the index dependent residuals."""

        self._reduced_clps = {}
        self._weighted_residuals = {}
        self._residuals = {}

        for label, problem in self.bag.items():
            self._calculate_residual_for_problem(label, problem)

        self._clps = (
            self.model.retrieve_clp_function(
                self.parameters,
                self.clp_labels,
                self.reduced_clp_labels,
                self.reduced_clps,
                self.data,
            )
            if callable(self.model.retrieve_clp_function)
            else self.reduced_clps
        )

        return self._reduced_clps, self._clps, self._weighted_residuals, self._residuals

    def calculate_index_independent_residual(
        self,
    ) -> tuple[
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the index independent residuals."""
        return self.calculate_index_dependent_residual()

    def _calculate_residual_for_problem(self, label: str, problem: UngroupedProblemDescriptor):
        self._reduced_clps[label] = []
        self._weighted_residuals[label] = []
        self._residuals[label] = []
        data = problem.data

        for i in range(len(problem.global_axis)):
            matrix = (
                self.reduced_matrices[label][i]
                if self.index_dependent
                else self.reduced_matrices[label].copy()
            )  # TODO: .copy() or not
            if problem.dataset.scale is not None:
                matrix *= self.filled_dataset_descriptors[label].scale

            if problem.weight is not None:
                for j in range(matrix.shape[1]):
                    matrix[:, j] *= problem.weight.isel({self._global_dimension: i}).values

            clp, residual = self._residual_function(
                matrix, data.isel({self._global_dimension: i}).values
            )
            self._reduced_clps[label].append(clp)
            self._weighted_residuals[label].append(residual)
            if problem.weight is not None:
                self._residuals[label].append(
                    residual / problem.weight.isel({self._global_dimension: i})
                )
            else:
                self._residuals[label].append(residual)

    def create_index_dependent_result_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        """Creates a result datasets for index dependent matrices."""

        self._add_index_dependent_matrix_to_dataset(label, dataset)

        self._add_residual_and_full_clp_to_dataset(label, dataset)

        return dataset

    def create_index_independent_result_dataset(
        self, label: str, dataset: xr.Dataset
    ) -> xr.Dataset:
        """Creates a result datasets for index independent matrices."""

        self._add_index_independent_matrix_to_dataset(label, dataset)

        self._add_residual_and_full_clp_to_dataset(label, dataset)

        return dataset

    def _add_index_dependent_matrix_to_dataset(self, label: str, dataset: xr.Dataset):
        # we assume that the labels are the same, this might not be true in
        # future models
        dataset.coords["clp_label"] = self.clp_labels[label][0]

        dataset["matrix"] = (
            (
                (self._global_dimension),
                (self._model_dimension),
                ("clp_label"),
            ),
            np.asarray(self.matrices[label]),
        )

    def _add_index_independent_matrix_to_dataset(self, label: str, dataset: xr.Dataset):
        dataset.coords["clp_label"] = self.clp_labels[label]
        dataset["matrix"] = (
            (
                (self._model_dimension),
                ("clp_label"),
            ),
            np.asarray(self.matrices[label]),
        )

    def _add_residual_and_full_clp_to_dataset(self, label: str, dataset: xr.Dataset):
        dataset["clp"] = (
            (
                (self._global_dimension),
                ("clp_label"),
            ),
            np.asarray(self.clps[label]),
        )
        dataset["weighted_residual"] = (
            (
                (self._model_dimension),
                (self._global_dimension),
            ),
            np.transpose(np.asarray(self.weighted_residuals[label])),
        )
        dataset["residual"] = (
            (
                (self._model_dimension),
                (self._global_dimension),
            ),
            np.transpose(np.asarray(self.residuals[label])),
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.builtin.models.kinetic_image.kinetic_baseline_megacomplex import (
    KineticBaselineMegacomplex,
)
from glotaran.builtin.models.kinetic_image.kinetic_decay_megacomplex import KineticDecayMegacomplex
from glotaran.builtin.models.kinetic_image.kinetic_image_model import KineticImageModel
from glotaran.builtin.models.kinetic_spectrum.coherent_artifact_megacomplex import (
    CoherentArtifactMegacomplex,
)
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_dataset_descriptor import (
    KineticSpectrumDatasetDescriptor,
)
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_result import (
    finalize_kinetic_spectrum_result,
)
from glotaran.builtin.models.kinetic_spectrum.spectral_constraints import SpectralConstraint
from glotaran.builtin.models.kinetic_spectrum.spectral_constraints import (
    apply_spectral_constraints,
)
from glotaran.builtin.models.kinetic_spectrum.spectral_irf import IrfSpectralMultiGaussian
from glotaran.builtin.models.kinetic_spectrum.spectral_matrix import spectral_matrix
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import EqualAreaPenalty
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import apply_spectral_penalties
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import has_spectral_penalties
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import SpectralRelation
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import apply_spectral_relations
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import retrieve_related_clps
from glotaran.builtin.models.kinetic_spectrum.spectral_shape import SpectralShape
from glotaran.model import model

if TYPE_CHECKING:
    from glotaran.parameter import ParameterGroup


def has_kinetic_model_constraints(model: KineticSpectrumModel) -> bool:
    return len(model.spectral_relations) + len(model.spectral_constraints) != 0


def apply_kinetic_model_constraints(
    model: KineticSpectrumModel,
    dataset: str,
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
) -> tuple[list[str], np.ndarray]:
    clp_labels, matrix = apply_spectral_relations(
        model, dataset, parameters, clp_labels, matrix, index
    )
    clp_labels, matrix = apply_spectral_constraints(model, clp_labels, matrix, index)
    return clp_labels, matrix


def retrieve_spectral_clps(
    model: KineticSpectrumModel,
    parameters: ParameterGroup,
    clp_labels: dict[str, list[str] | list[list[str]]],
    reduced_clp_labels: dict[str, list[str] | list[list[str]]],
    reduced_clps: dict[str, list[np.ndarray]],
    data: dict[str, xr.Dataset],
) -> dict[str, list[np.ndarray]]:
    if not has_kinetic_model_constraints(model):
        return reduced_clps

    # Note: we are always in index_dependent case when we have constraints
    clps = {}
    for label in clp_labels:
        clps[label] = []
        for i, index_reduced_clp_labels in enumerate(reduced_clp_labels[label]):
            index_clp_labels = clp_labels[label][i]
            index_reduced_clps = reduced_clps[label][i]
            index_clps = np.zeros((len(index_clp_labels)), dtype=np.float64)
            for j, clp_label in enumerate(index_reduced_clp_labels):
                index_clps[index_clp_labels.index(clp_label)] = index_reduced_clps[j]
            clps[label].append(index_clps)
    clps = retrieve_related_clps(model, parameters, clp_labels, clps, data)
    return clps


def index_dependent(model: KineticSpectrumModel) -> bool:
    return (
        any(
            isinstance(irf, IrfSpectralMultiGaussian) and irf.dispersion_center is not None
            for irf in model.irf.values()
        )
        or len(model.spectral_relations) != 0
        or len(model.spectral_constraints) != 0
        or len(model.weights) != 0
    )


def grouped(model: KineticSpectrumModel):
    return len(model.dataset) != 1


@model(
    "kinetic-spectrum",
    attributes={
        "equal_area_penalties": EqualAreaPenalty,
        "shape": SpectralShape,
        "spectral_constraints": SpectralConstraint,
        "spectral_relations": SpectralRelation,
    },
    dataset_type=KineticSpectrumDatasetDescriptor,
    default_megacomplex_type="kinetic-decay",
    megacomplex_types={
        "coherent-artifact": CoherentArtifactMegacomplex,
        "kinetic-decay": KineticDecayMegacomplex,
        "kinetic-baseline": KineticBaselineMegacomplex,
    },
    model_dimension="time",
    global_matrix=spectral_matrix,
    global_dimension="spectral",
    has_matrix_constraints_function=has_kinetic_model_constraints,
    constrain_matrix_function=apply_kinetic_model_constraints,
    retrieve_clp_function=retrieve_spectral_clps,
    has_additional_penalty_function=has_spectral_penalties,
    additional_penalty_function=apply_spectral_penalties,
    grouped=grouped,
    index_dependent=index_dependent,
    finalize_data_function=finalize_kinetic_spectrum_result,
)
class KineticSpectrumModel(KineticImageModel):
    pass



"""This package contains the kinetic megacomplex item."""
from __future__ import annotations

import numba as nb
import numpy as np

from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.model import DatasetDescriptor
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import model_attribute
from glotaran.parameter import Parameter


@model_attribute(
    properties={
        "order": {"type": int},
        "width": {"type": Parameter, "allow_none": True},
    },
)
class CoherentArtifactMegacomplex(Megacomplex):
    def calculate_matrix(
        self,
        model,
        dataset_descriptor: DatasetDescriptor,
        indices: dict[str, int],
        axis: dict[str, np.ndarray],
        **kwargs,
    ):
        if not 1 <= self.order <= 3:
            raise ModelError("Coherent artifact order must be between in [1,3]")

        if dataset_descriptor.irf is None:
            raise ModelError(f'No irf in dataset "{dataset_descriptor.label}"')

        if not isinstance(dataset_descriptor.irf, IrfMultiGaussian):
            raise ModelError(f'Irf in dataset "{dataset_descriptor.label} is not a gaussian irf."')

        global_index = indices.get(model.global_dimension, None)
        global_axis = axis.get(model.global_dimension, None)
        irf = dataset_descriptor.irf

        center, width, _, _, _, _ = irf.parameter(global_index, global_axis)
        center = center[0]
        width = self.width.value if self.width is not None else width[0]

        axis = axis[model.model_dimension]
        matrix = _calculate_coherent_artifact_matrix(center, width, axis, self.order)
        return (self.compartments(), matrix)

    def compartments(self):
        return [f"coherent_artifact_{i}" for i in range(1, self.order + 1)]


@nb.jit(nopython=True, parallel=True)
def _calculate_coherent_artifact_matrix(center, width, axis, order):
    matrix = np.zeros((axis.size, order), dtype=np.float64)

    matrix[:, 0] = np.exp(-1 * (axis - center) ** 2 / (2 * width ** 2))
    if order > 1:
        matrix[:, 1] = matrix[:, 0] * (center - axis) / width ** 2

    if order > 2:
        matrix[:, 2] = (
            matrix[:, 0] * (center ** 2 - width ** 2 - 2 * center * axis + axis ** 2) / width ** 4
        )
    return matrix



from __future__ import annotations

from glotaran.builtin.models.kinetic_image.kinetic_image_dataset_descriptor import (
    KineticImageDatasetDescriptor,
)
from glotaran.model import model_attribute

class KineticSpectrumDatasetDescriptor(KineticImageDatasetDescriptor):
    @property
    def shape(self) -> dict[str, str]: ...



from __future__ import annotations

from typing import Any
from typing import Mapping

import numpy as np

from glotaran.builtin.models.kinetic_image.kinetic_image_megacomplex import KineticImageMegacomplex
from glotaran.builtin.models.kinetic_image.kinetic_image_model import KineticImageModel
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_dataset_descriptor import (
    KineticSpectrumDatasetDescriptor,
)
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_matrix import (
    kinetic_spectrum_matrix,
)
from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_result import (
    finalize_kinetic_spectrum_result,
)
from glotaran.builtin.models.kinetic_spectrum.spectral_constraints import SpectralConstraint
from glotaran.builtin.models.kinetic_spectrum.spectral_constraints import (
    apply_spectral_constraints,
)
from glotaran.builtin.models.kinetic_spectrum.spectral_irf import IrfSpectralMultiGaussian
from glotaran.builtin.models.kinetic_spectrum.spectral_matrix import spectral_matrix
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import EqualAreaPenalty
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import apply_spectral_penalties
from glotaran.builtin.models.kinetic_spectrum.spectral_penalties import has_spectral_penalties
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import SpectralRelation
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import apply_spectral_relations
from glotaran.builtin.models.kinetic_spectrum.spectral_relations import retrieve_related_clps
from glotaran.builtin.models.kinetic_spectrum.spectral_shape import SpectralShape
from glotaran.model import model
from glotaran.parameter import ParameterGroup

def has_kinetic_model_constraints(model: KineticSpectrumModel) -> bool: ...  # noqa: F811
def apply_kinetic_model_constraints(
    model: KineticSpectrumModel,  # noqa: F811
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
) -> Any: ...
def retrieve_spectral_clps(
    model: KineticSpectrumModel,  # noqa: F811
    parameters: ParameterGroup,
    clp_labels: list[str],
    reduced_clp_labels: list[str],
    reduced_clps: np.ndarray | list[np.ndarray],
    global_axis: np.ndarray,
) -> Any: ...
def index_dependent(model: KineticSpectrumModel) -> Any: ...  # noqa: F811
def grouped(model: KineticSpectrumModel) -> bool: ...  # noqa: F811

class KineticSpectrumModel(KineticImageModel):
    dataset: Mapping[str, KineticSpectrumDatasetDescriptor]
    megacomplex: Mapping[str, KineticImageMegacomplex]
    @property
    def equal_area_penalties(self) -> list[EqualAreaPenalty]: ...
    @property
    def shape(self) -> Mapping[str, SpectralShape]: ...
    @property
    def spectral_constraints(self) -> list[SpectralConstraint]: ...
    @property
    def spectral_relations(self) -> list[SpectralRelation]: ...
    def has_matrix_constraints_function(self) -> bool: ...
    def constrain_matrix_function(
        self, parameters: ParameterGroup, clp_labels: list[str], matrix: np.ndarray, index: float
    ) -> tuple[list[str], np.ndarray]: ...
    def retrieve_clp_function(
        self,
        parameters: ParameterGroup,
        clp_labels: list[str],
        reduced_clp_labels: list[str],
        reduced_clps: np.ndarray | list[np.ndarray],
        global_axis: np.ndarray,
    ) -> np.ndarray | list[np.ndarray]: ...
    def has_additional_penalty_function(self) -> bool: ...
    def additional_penalty_function(
        self,
        parameters: ParameterGroup,
        clp_labels: list[str] | list[list[str]],
        clps: np.ndarray,
        global_axis: np.ndarray,
    ) -> np.ndarray: ...
    @staticmethod
    def global_matrix(dataset, axis) -> tuple[None, None] | tuple[list[str], np.ndarray]: ...
    @staticmethod
    def matrix(  # type: ignore[override]
        dataset_descriptor: KineticSpectrumDatasetDescriptor = ...,
        axis=...,
        index=...,
        irf=...,
    ) -> tuple[None, None] | tuple[list[Any], np.ndarray]: ...



from glotaran.builtin.models.kinetic_image.kinetic_image_dataset_descriptor import (
    KineticImageDatasetDescriptor,
)
from glotaran.builtin.models.spectral.shape import SpectralShape
from glotaran.builtin.models.spectral.spectral_megacomplex import SpectralMegacomplex
from glotaran.builtin.models.spectral.spectral_result import finalize_spectral_result
from glotaran.model import Model
from glotaran.model import model


@model(
    "spectral-model",
    attributes={
        "shape": SpectralShape,
    },
    dataset_type=KineticImageDatasetDescriptor,
    default_megacomplex_type="spectral",
    megacomplex_types={
        "spectral": SpectralMegacomplex,
    },
    model_dimension="spectral",
    global_dimension="time",
    grouped=False,
    index_dependent=False,
    finalize_data_function=finalize_spectral_result,
)
class SpectralModel(Model):
    pass



"""Model registration convenience functions."""
from __future__ import annotations

from typing import TYPE_CHECKING

from tabulate import tabulate

from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.base_registry import add_plugin_to_registry
from glotaran.plugin_system.base_registry import full_plugin_name
from glotaran.plugin_system.base_registry import get_plugin_from_registry
from glotaran.plugin_system.base_registry import is_registered_plugin
from glotaran.plugin_system.base_registry import registered_plugins
from glotaran.plugin_system.base_registry import set_plugin
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from glotaran.model import Model


def register_model(model_type: str, model: type[Model]) -> None:
    """Add a model to the model registry.

    Parameters
    ----------
    model_type : str
        Name of the model under which it is registered.
    model : type[Model]
        model class to be registered.
    """
    add_plugin_to_registry(
        plugin_register_key=model_type,
        plugin=model,
        plugin_registry=__PluginRegistry.model,
        plugin_set_func_name="set_model_plugin",
    )


def is_known_model(model_type: str) -> bool:
    """Check if a model is in the model registry.

    Parameters
    ----------
    model_type : str
        Name of the model under which it is registered.

    Returns
    -------
    bool
        Whether or not the model is registered.
    """
    return is_registered_plugin(
        plugin_register_key=model_type, plugin_registry=__PluginRegistry.model
    )


def get_model(model_type: str) -> type[Model]:
    """Retrieve a model from the model registry.

    Parameters
    ----------
    model_type : str
        Name of the model under which it is registered.

    Returns
    -------
    type[Model]
        Model class
    """
    return get_plugin_from_registry(
        plugin_register_key=model_type,
        plugin_registry=__PluginRegistry.model,
        not_found_error_message=(
            f"Unknown model type {model_type!r}. "
            f"Known model types are: {known_model_names(full_names=True)}"
        ),
    )


def known_model_names(full_names: bool = False) -> list[str]:
    """Names of the registered models.

    Parameters
    ----------
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    list[str]
        List of registered models.
    """
    return registered_plugins(__PluginRegistry.model, full_names=full_names)


def set_model_plugin(model_name: str, full_plugin_name: str) -> None:
    """Set the plugin used for a specific model name.

    This function is useful when you want to resolve conflicts of installed plugins
    or overwrite the plugin used for a specific model name.

    Effected functions:

    - :func:`optimize`

    Parameters
    ----------
    model_name : str
        Name of the model to use the plugin for.
    full_plugin_name : str
        Full name (import path) of the registered plugin.
    """
    set_plugin(
        plugin_register_key=model_name,
        full_plugin_name=full_plugin_name,
        plugin_registry=__PluginRegistry.model,
        plugin_register_key_name="model_name",
    )


def model_plugin_table(*, plugin_names: bool = False, full_names: bool = False) -> MarkdownStr:
    """Return registered model plugins as markdown table.

    This is especially useful when you work with new plugins.

    Parameters
    ----------
    plugin_names : bool
        Whether or not to add the names of the plugins to the table.
    full_names : bool
        Whether to display the full names the plugins are
        registered under as well.

    Returns
    -------
    str
        Markdown table of modelnames.
    """
    table_data = []
    model_names = known_model_names(full_names=full_names)
    header_values = ["Model name"]
    if plugin_names:
        header_values.append("Plugin name")
        for model_name in model_names:
            table_data.append([f"`{model_name}`", f"`{full_plugin_name(get_model(model_name))}`"])
    else:
        table_data = [[f"`{model_name}`"] for model_name in model_names]
    headers = tuple(map(lambda x: f"__{x}__", header_values))
    return MarkdownStr(tabulate(table_data, tablefmt="github", headers=headers, stralign="center"))



"""Glotaran Model Package

This package contains the Glotaran's base model object, the model decorators and
common model items.
"""

from glotaran.model.attribute import model_attribute
from glotaran.model.attribute import model_attribute_typed
from glotaran.model.base_model import Model
from glotaran.model.dataset_descriptor import DatasetDescriptor
from glotaran.model.decorator import model
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.util import ModelError
from glotaran.model.weight import Weight
from glotaran.plugin_system.model_registration import get_model
from glotaran.plugin_system.model_registration import is_known_model
from glotaran.plugin_system.model_registration import known_model_names
from glotaran.plugin_system.model_registration import model_plugin_table
from glotaran.plugin_system.model_registration import set_model_plugin



from __future__ import annotations

from typing import TypeVar

from glotaran.model.attribute import model_attribute  # noqa: F401
from glotaran.model.base_model import Model
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup

_T_Model = TypeVar("_T_Model", bound=Model)
_T_DatasetDescriptor = TypeVar("_T_DatasetDescriptor", bound=DatasetDescriptor)  # noqa: F821

class DatasetDescriptor:
    megacomplex: list[str]
    megacomplex_scale: list[Parameter] | None = ...
    scale: Parameter | None = ...
    def fill(self, model: _T_Model, parameters: ParameterGroup) -> _T_DatasetDescriptor: ...
    @classmethod
    def from_dict(cls: type[_T_DatasetDescriptor], values: dict) -> _T_DatasetDescriptor: ...
    @classmethod
    def from_list(cls: type[_T_DatasetDescriptor], values: list) -> _T_DatasetDescriptor: ...
    def validate(self, model: _T_Model, parameters=...) -> list[str]: ...
    def mprint_item(
        self, parameters: ParameterGroup = ..., initial_parameters: ParameterGroup = ...
    ) -> str: ...



from __future__ import annotations

import xarray as xr

from glotaran.analysis.problem import Problem
from glotaran.builtin.models.spectral.spectral_megacomplex import SpectralMegacomplex


def finalize_spectral_result(model, problem: Problem, data: dict[str, xr.Dataset]):

    for label, dataset in data.items():

        dataset_descriptor = problem.filled_dataset_descriptors[label]

        retrieve_spectral_data(problem.model, dataset, dataset_descriptor)


def retrieve_spectral_data(model, dataset, dataset_descriptor):
    spectral_compartments = []
    for megacomplex in dataset_descriptor.megacomplex:
        if isinstance(megacomplex, SpectralMegacomplex):
            spectral_compartments += [
                compartment
                for compartment in megacomplex.shape
                if compartment not in spectral_compartments
            ]

    dataset.coords["species"] = spectral_compartments
    dataset["species_spectra"] = (
        (
            model.model_dimension,
            "species",
        ),
        dataset.matrix.sel(clp_label=spectral_compartments).values,
    )
    dataset["species_associated_concentrations"] = (
        (
            model.global_dimension,
            "species",
        ),
        dataset.clp.sel(clp_label=spectral_compartments).data,
    )



from __future__ import annotations

import itertools
from typing import NamedTuple

import numpy as np

from glotaran.model import DatasetDescriptor
from glotaran.model import Model
from glotaran.parameter import ParameterGroup


class LabelAndMatrix(NamedTuple):
    clp_label: list[str]
    matrix: np.ndarray


def find_overlap(a, b, rtol=1e-05, atol=1e-08):
    ovr_a = []
    ovr_b = []
    start_b = 0
    for i, ai in enumerate(a):
        for j, bj in itertools.islice(enumerate(b), start_b, None):
            if np.isclose(ai, bj, rtol=rtol, atol=atol, equal_nan=False):
                ovr_a.append(i)
                ovr_b.append(j)
            elif bj > ai:  # (more than tolerance)
                break  # all the rest will be farther away
            else:  # bj < ai (more than tolerance)
                start_b += 1  # ignore further tests of this item
    return (ovr_a, ovr_b)


def find_closest_index(index: float, axis: np.ndarray):
    return np.abs(axis - index).argmin()


def get_min_max_from_interval(interval, axis):
    minimum = np.abs(axis.values - interval[0]).argmin() if not np.isinf(interval[0]) else 0
    maximum = (
        np.abs(axis.values - interval[1]).argmin() + 1 if not np.isinf(interval[1]) else axis.size
    )
    return slice(minimum, maximum)


def calculate_matrix(
    model: Model,
    dataset_descriptor: DatasetDescriptor,
    indices: dict[str, int],
    axis: dict[str, np.ndarray],
) -> LabelAndMatrix:
    clp_labels = None
    matrix = None

    for scale, megacomplex in dataset_descriptor.iterate_megacomplexes():
        this_clp_labels, this_matrix = megacomplex.calculate_matrix(
            model, dataset_descriptor, indices, axis
        )

        if scale is not None:
            this_matrix *= scale

        if matrix is None:
            clp_labels = this_clp_labels
            matrix = this_matrix
        else:
            tmp_clp_labels = clp_labels + [c for c in this_clp_labels if c not in clp_labels]
            tmp_matrix = np.zeros((matrix.shape[0], len(tmp_clp_labels)), dtype=np.float64)
            for idx, label in enumerate(tmp_clp_labels):
                if label in clp_labels:
                    tmp_matrix[:, idx] += matrix[:, clp_labels.index(label)]
                if label in this_clp_labels:
                    tmp_matrix[:, idx] += this_matrix[:, this_clp_labels.index(label)]
            clp_labels = tmp_clp_labels
            matrix = tmp_matrix

    return LabelAndMatrix(clp_labels, matrix)


def reduce_matrix(
    model: Model,
    label: str,
    parameters: ParameterGroup,
    result: LabelAndMatrix,
    index: float | None,
) -> LabelAndMatrix:
    clp_labels = result.clp_label.copy()
    if callable(model.has_matrix_constraints_function) and model.has_matrix_constraints_function():
        clp_label, matrix = model.constrain_matrix_function(
            label, parameters, clp_labels, result.matrix, index
        )
        return LabelAndMatrix(clp_label, matrix)
    return LabelAndMatrix(clp_labels, result.matrix)


def combine_matrices(labels_and_matrices: list[LabelAndMatrix]) -> LabelAndMatrix:
    masks = []
    full_clp_labels = None
    sizes = []
    for label_and_matrix in labels_and_matrices:
        (clp_label, matrix) = label_and_matrix
        sizes.append(matrix.shape[0])
        if full_clp_labels is None:
            full_clp_labels = clp_label
            masks.append([i for i, _ in enumerate(clp_label)])
        else:
            mask = []
            for c in clp_label:
                if c not in full_clp_labels:
                    full_clp_labels.append(c)
                mask.append(full_clp_labels.index(c))
            masks.append(mask)
    dim1 = np.sum(sizes)
    dim2 = len(full_clp_labels)
    full_matrix = np.zeros((dim1, dim2), dtype=np.float64)
    start = 0
    for i, m in enumerate(labels_and_matrices):
        end = start + sizes[i]
        full_matrix[start:end, masks[i]] = m[1]
        start = end

    return LabelAndMatrix(full_clp_labels, full_matrix)



from __future__ import annotations

import collections
import itertools
from typing import Deque

import numpy as np
import xarray as xr

from glotaran.analysis.problem import GroupedProblemDescriptor
from glotaran.analysis.problem import Problem
from glotaran.analysis.problem import ProblemGroup
from glotaran.analysis.util import LabelAndMatrix
from glotaran.analysis.util import calculate_matrix
from glotaran.analysis.util import combine_matrices
from glotaran.analysis.util import find_closest_index
from glotaran.analysis.util import find_overlap
from glotaran.analysis.util import reduce_matrix
from glotaran.model import DatasetDescriptor


class GroupedProblem(Problem):
    """Represents a problem where the data is grouped."""

    def init_bag(self):
        """Initializes a grouped problem bag."""
        datasets = None
        for label in self._model.dataset:
            dataset = self._data[label]
            if "weight" in dataset:
                weight = dataset.weight
                data = dataset.data * weight
                dataset["weighted_data"] = data
            else:
                weight = xr.DataArray(np.ones_like(dataset.data), coords=dataset.data.coords)
                data = dataset.data
            global_axis = dataset.coords[self._global_dimension].values
            model_axis = dataset.coords[self._model_dimension].values
            has_scaling = self._model.dataset[label].scale is not None
            if self._bag is None:
                self._bag = collections.deque(
                    ProblemGroup(
                        data=data.isel({self._global_dimension: i}).values,
                        weight=weight.isel({self._global_dimension: i}).values,
                        has_scaling=has_scaling,
                        group=label,
                        data_sizes=[data.isel({self._global_dimension: i}).values.size],
                        descriptor=[
                            GroupedProblemDescriptor(
                                label,
                                {
                                    self._global_dimension: i,
                                },
                                {
                                    self._model_dimension: model_axis,
                                    self._global_dimension: global_axis,
                                },
                            )
                        ],
                    )
                    for i, value in enumerate(global_axis)
                )
                datasets = collections.deque([label] for _, _ in enumerate(global_axis))
                self._full_axis = collections.deque(global_axis)
            else:
                self._append_to_grouped_bag(
                    label, datasets, global_axis, model_axis, data, weight, has_scaling
                )
        self._full_axis = np.asarray(self._full_axis)
        self._groups = {"".join(d): d for d in datasets}

    def _append_to_grouped_bag(
        self,
        label: str,
        datasets: Deque[str],
        global_axis: np.ndarray,
        model_axis: np.ndarray,
        data: xr.DataArray,
        weight: xr.DataArray,
        has_scaling: bool,
    ):
        i1, i2 = find_overlap(self._full_axis, global_axis, atol=self._scheme.group_tolerance)

        for i, j in enumerate(i1):
            datasets[j].append(label)
            data_stripe = data.isel({self._global_dimension: i2[i]}).values
            self._bag[j] = ProblemGroup(
                data=np.concatenate(
                    [
                        self._bag[j].data,
                        data_stripe,
                    ]
                ),
                weight=np.concatenate(
                    [
                        self._bag[j].weight,
                        weight.isel({self._global_dimension: i2[i]}).values,
                    ]
                ),
                has_scaling=has_scaling or self._bag[j].has_scaling,
                group=self._bag[j].group + label,
                data_sizes=self._bag[j].data_sizes + [data_stripe.size],
                descriptor=self._bag[j].descriptor
                + [
                    GroupedProblemDescriptor(
                        label,
                        {
                            self._global_dimension: i2[i],
                        },
                        {
                            self._model_dimension: model_axis,
                            self._global_dimension: global_axis,
                        },
                    )
                ],
            )

        # Add non-overlaping regions
        begin_overlap = i2[0] if len(i2) != 0 else 0
        end_overlap = i2[-1] + 1 if len(i2) != 0 else 0
        for i in itertools.chain(range(begin_overlap), range(end_overlap, len(global_axis))):
            data_stripe = data.isel({self._global_dimension: i}).values
            problem = ProblemGroup(
                data=data_stripe,
                weight=weight.isel({self._global_dimension: i}).values,
                has_scaling=has_scaling,
                group=label,
                data_sizes=[data_stripe.size],
                descriptor=[
                    GroupedProblemDescriptor(
                        label,
                        {
                            self._global_dimension: i,
                        },
                        {
                            self._model_dimension: model_axis,
                            self._global_dimension: global_axis,
                        },
                    )
                ],
            )
            if i < end_overlap:
                datasets.appendleft([label])
                self._full_axis.appendleft(global_axis[i])
                self._bag.appendleft(problem)
            else:
                datasets.append([label])
                self._full_axis.append(global_axis[i])
                self._bag.append(problem)

    def calculate_index_dependent_matrices(
        self,
    ) -> tuple[
        dict[str, list[list[str]]],
        dict[str, list[np.ndarray]],
        list[list[str]],
        list[np.ndarray],
    ]:
        """Calculates the index dependent model matrices."""

        def calculate_group(
            group: ProblemGroup, descriptors: dict[str, DatasetDescriptor]
        ) -> tuple[list[tuple[LabelAndMatrix, str]], float]:
            result = [
                (
                    calculate_matrix(
                        self._model,
                        descriptors[problem.label],
                        problem.indices,
                        problem.axis,
                    ),
                    problem.label,
                )
                for problem in group.descriptor
            ]
            global_index = group.descriptor[0].indices[self._global_dimension]
            global_index = group.descriptor[0].axis[self._global_dimension][global_index]
            return result, global_index

        def reduce_and_combine_matrices(
            results: tuple[list[tuple[LabelAndMatrix, str]], float],
        ) -> LabelAndMatrix:
            index_results, index = results
            constraint_labels_and_matrices = list(
                map(
                    lambda result: reduce_matrix(
                        self._model, result[1], self.parameters, result[0], index
                    ),
                    index_results,
                )
            )
            clp, matrix = combine_matrices(constraint_labels_and_matrices)
            return LabelAndMatrix(clp, matrix)

        results = list(
            map(lambda group: calculate_group(group, self._filled_dataset_descriptors), self._bag)
        )

        clp_labels = list(map(lambda result: [r[0].clp_label for r in result[0]], results))
        matrices = list(map(lambda result: [r[0].matrix for r in result[0]], results))

        self._clp_labels = {}
        self._matrices = {}

        for i, grouped_problem in enumerate(self._bag):
            for j, descriptor in enumerate(grouped_problem.descriptor):
                if descriptor.label not in self._clp_labels:
                    self._clp_labels[descriptor.label] = []
                    self._matrices[descriptor.label] = []
                self._clp_labels[descriptor.label].append(clp_labels[i][j])
                self._matrices[descriptor.label].append(matrices[i][j])

        reduced_results = list(map(reduce_and_combine_matrices, results))
        self._reduced_clp_labels = list(map(lambda result: result.clp_label, reduced_results))
        self._reduced_matrices = list(map(lambda result: result.matrix, reduced_results))
        return self._clp_labels, self._matrices, self._reduced_clp_labels, self._reduced_matrices

    def calculate_index_independent_matrices(
        self,
    ) -> tuple[dict[str, list[str]], dict[str, np.ndarray], dict[str, LabelAndMatrix],]:
        """Calculates the index independent model matrices."""
        self._clp_labels = {}
        self._matrices = {}
        self._reduced_clp_labels = {}
        self._reduced_matrices = {}

        for label, descriptor in self._filled_dataset_descriptors.items():
            model_axis = self._data[label].coords[self._model_dimension].values
            global_axis = self._data[label].coords[self._global_dimension].values
            result = calculate_matrix(
                self._model,
                descriptor,
                {},
                {
                    self._model_dimension: model_axis,
                    self._global_dimension: global_axis,
                },
            )

            self._clp_labels[label] = result.clp_label
            self._matrices[label] = result.matrix
            reduced_result = reduce_matrix(self._model, label, self._parameters, result, None)
            self._reduced_clp_labels[label] = reduced_result.clp_label
            self._reduced_matrices[label] = reduced_result.matrix

        for group_label, group in self.groups.items():
            if group_label not in self._matrices:
                reduced_labels_and_matrix = combine_matrices(
                    [
                        LabelAndMatrix(
                            self._reduced_clp_labels[label], self._reduced_matrices[label]
                        )
                        for label in group
                    ]
                )
                self._reduced_clp_labels[group_label] = reduced_labels_and_matrix.clp_label
                self._reduced_matrices[group_label] = reduced_labels_and_matrix.matrix

        return self._clp_labels, self._matrices, self._reduced_clp_labels, self._reduced_matrices

    def calculate_index_dependent_residual(
        self,
    ) -> tuple[list[np.ndarray], list[np.ndarray], list[np.ndarray], list[np.ndarray],]:
        """Calculates the index dependent residuals."""

        def residual_function(
            problem: ProblemGroup, matrix: np.ndarray
        ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:

            matrix = matrix.copy()
            for i in range(matrix.shape[1]):
                matrix[:, i] *= problem.weight
            data = problem.data
            if problem.has_scaling:
                for i, descriptor in enumerate(problem.descriptor):
                    label = descriptor.label
                    if self.filled_dataset_descriptors[label] is not None:
                        start = sum(problem.data_sizes[0:i])
                        end = start + problem.data_sizes[i]
                        matrix[start:end, :] *= self.filled_dataset_descriptors[label].scale

            clp, residual = self._residual_function(matrix, data)
            return clp, residual, residual / problem.weight

        results = list(map(residual_function, self.bag, self.reduced_matrices))

        self._weighted_residuals = list(map(lambda result: result[1], results))
        self._residuals = list(map(lambda result: result[2], results))

        reduced_clps = list(map(lambda result: result[0], results))
        self._ungroup_clps(reduced_clps)

        return self._reduced_clps, self._clps, self._weighted_residuals, self._residuals

    def calculate_index_independent_residual(
        self,
    ) -> tuple[list[np.ndarray], list[np.ndarray], list[np.ndarray], list[np.ndarray],]:
        """Calculates the index independent residuals."""

        def residual_function(problem: ProblemGroup):
            matrix = self.reduced_matrices[problem.group].copy()
            for i in range(matrix.shape[1]):
                matrix[:, i] *= problem.weight
            data = problem.data
            if problem.has_scaling:
                for i, descriptor in enumerate(problem.descriptor):
                    label = descriptor.label
                    if self.filled_dataset_descriptors[label] is not None:
                        start = sum(problem.data_sizes[0:i])
                        end = start + problem.data_sizes[i]
                        matrix[start:end, :] *= self.filled_dataset_descriptors[label].scale
            clp, residual = self._residual_function(matrix, data)
            return clp, residual, residual / problem.weight

        results = list(map(residual_function, self.bag))

        self._weighted_residuals = list(map(lambda result: result[1], results))
        self._residuals = list(map(lambda result: result[2], results))

        reduced_clps = list(map(lambda result: result[0], results))
        self._ungroup_clps(reduced_clps)

        return self._reduced_clps, self._clps, self._weighted_residuals, self._residuals

    def _ungroup_clps(self, reduced_clps: np.ndarray):
        reduced_clp_labels = self.reduced_clp_labels
        self._reduced_clp_labels = {}
        self._reduced_clps = {}
        for label, clp_labels in self.clp_labels.items():

            # find offset in the full axis
            offset = find_closest_index(
                self.data[label].coords[self._global_dimension][0].values, self._full_axis
            )

            self._reduced_clp_labels[label] = []
            self._reduced_clps[label] = []
            for i in range(self.data[label].coords[self._global_dimension].size):
                group_label = self.bag[i].group
                dataset_clp_labels = clp_labels[i] if self._index_dependent else clp_labels
                index_clp_labels = (
                    reduced_clp_labels[i + offset]
                    if self._index_dependent
                    else reduced_clp_labels[group_label]
                )
                self._reduced_clp_labels[label].append(
                    [
                        clp_label
                        for clp_label in dataset_clp_labels
                        if clp_label in index_clp_labels
                    ]
                )

                mask = [
                    clp_label in self._reduced_clp_labels[label][i]
                    for clp_label in index_clp_labels
                ]
                self._reduced_clps[label].append(reduced_clps[i + offset][mask])
        self._clps = (
            self.model.retrieve_clp_function(
                self.parameters,
                self.clp_labels,
                self.reduced_clp_labels,
                self.reduced_clps,
                self.data,
            )
            if callable(self.model.retrieve_clp_function)
            else self._reduced_clps
        )

    def create_index_dependent_result_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        """Creates a result datasets for index dependent matrices."""

        for index, grouped_problem in enumerate(self.bag):

            if label in grouped_problem.group:
                group_index = [
                    descriptor.label for descriptor in grouped_problem.descriptor
                ].index(label)
                group_descriptor = grouped_problem.descriptor[group_index]
                global_index = group_descriptor.indices[self._global_dimension]
                global_index = group_descriptor.axis[self._global_dimension][global_index]

                self._add_grouped_residual_to_dataset(
                    dataset, grouped_problem, index, group_index, global_index
                )

        # we assume that the labels are the same, this might not be true in
        # future models
        dataset.coords["clp_label"] = self.clp_labels[label][0]
        dataset["matrix"] = (
            (
                (self._global_dimension),
                (self._model_dimension),
                ("clp_label"),
            ),
            self.matrices[label],
        )
        dataset["clp"] = (
            (
                (self._global_dimension),
                ("clp_label"),
            ),
            self.clps[label],
        )

        return dataset

    def create_index_independent_result_dataset(
        self, label: str, dataset: xr.Dataset
    ) -> xr.Dataset:
        """Creates a result datasets for index independent matrices."""

        self._add_index_independent_matrix_to_dataset(label, dataset)

        for index, grouped_problem in enumerate(self.bag):

            if label in grouped_problem.group:
                group_index = [
                    descriptor.label for descriptor in grouped_problem.descriptor
                ].index(label)
                group_descriptor = grouped_problem.descriptor[group_index]
                global_index = group_descriptor.indices[self._global_dimension]
                global_index = group_descriptor.axis[self._global_dimension][global_index]

                self._add_grouped_residual_to_dataset(
                    dataset, grouped_problem, index, group_index, global_index
                )

        dataset["clp"] = (
            (
                (self._global_dimension),
                ("clp_label"),
            ),
            self.clps[label],
        )

        return dataset

    def _add_index_independent_matrix_to_dataset(self, label: str, dataset: xr.Dataset):
        dataset.coords["clp_label"] = self.clp_labels[label]
        dataset["matrix"] = (
            (
                (self._model_dimension),
                ("clp_label"),
            ),
            np.asarray(self.matrices[label]),
        )

    def _add_grouped_residual_to_dataset(
        self,
        dataset: xr.Dataset,
        grouped_problem: ProblemGroup,
        index: int,
        group_index: int,
        global_index: int,
    ):
        if "residual" not in dataset:
            dim1 = dataset.coords[self._model_dimension].size
            dim2 = dataset.coords[self._global_dimension].size
            dataset["weighted_residual"] = (
                (self._model_dimension, self._global_dimension),
                np.zeros((dim1, dim2), dtype=np.float64),
            )
            dataset["residual"] = (
                (self._model_dimension, self._global_dimension),
                np.zeros((dim1, dim2), dtype=np.float64),
            )

        start = sum(
            self.data[grouped_problem.descriptor[i].label].coords[self._model_dimension].size
            for i in range(group_index)
        )

        end = start + dataset.coords[self._model_dimension].size
        dataset.weighted_residual.loc[
            {self._global_dimension: global_index}
        ] = self.weighted_residuals[index][start:end]
        dataset.residual.loc[{self._global_dimension: global_index}] = self.residuals[index][
            start:end
        ]



from __future__ import annotations

from typing import Any

import numpy as np

from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model import KineticSpectrumModel
from glotaran.model import model_attribute
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup

class SpectralRelation:
    @property
    def compartment(self) -> str: ...
    @property
    def target(self) -> str: ...
    @property
    def parameter(self) -> Parameter: ...
    @property
    def interval(self) -> list[tuple[float, float]]: ...
    def applies(self, index: Any) -> bool: ...

def create_spectral_relation_matrix(
    model: KineticSpectrumModel,
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
) -> tuple[list[str], np.ndarray]: ...
def apply_spectral_relations(
    model: KineticSpectrumModel,
    parameters: ParameterGroup,
    clp_labels: list[str],
    matrix: np.ndarray,
    index: float,
) -> tuple[list[str], np.ndarray]: ...
def retrieve_related_clps(
    model: KineticSpectrumModel,
    parameters: ParameterGroup,
    clp_labels: list[str],
    clps: np.ndarray,
    index: float,
) -> tuple[list[str], np.ndarray]: ...



import numpy as np
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.util import calculate_matrix
from glotaran.builtin.models.kinetic_spectrum import KineticSpectrumModel
from glotaran.builtin.models.kinetic_spectrum.spectral_constraints import (
    apply_spectral_constraints,
)
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


def test_spectral_constraint():
    model = KineticSpectrumModel.from_dict(
        {
            "initial_concentration": {
                "j1": {
                    "compartments": ["s1", "s2"],
                    "parameters": ["i.1", "i.2"],
                },
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s2", "s1"): "kinetic.1",
                        ("s2", "s2"): "kinetic.2",
                    }
                }
            },
            "spectral_constraints": [
                {"type": "zero", "compartment": "s2", "interval": (float("-inf"), float("inf"))},
            ],
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )
    print(model)

    wanted_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [1e-4, 1e-5],
            "i": [1, 2],
        }
    )
    initial_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [2e-4, 2e-5],
            "i": [1, 2, {"vary": False}],
        }
    )

    time = np.asarray(np.arange(0, 50, 1.5))
    dataset = model.dataset["dataset1"].fill(model, wanted_parameters)
    compartments, matrix = calculate_matrix(model, dataset, {}, {"time": time})

    assert len(compartments) == 2
    assert matrix.shape == (time.size, 2)

    reduced_compartments, reduced_matrix = apply_spectral_constraints(
        model, compartments, matrix, 1
    )

    print(reduced_matrix)
    assert len(reduced_compartments) == 1
    assert reduced_matrix.shape == (time.size, 1)

    reduced_compartments, reduced_matrix = model.constrain_matrix_function(
        "dataset1", wanted_parameters, compartments, matrix, 1
    )

    assert reduced_matrix.shape == (time.size, 1)

    clp = xr.DataArray(
        [[1.0, 10.0, 20.0, 1]], coords=(("spectral", [1]), ("clp_label", ["s1", "s2", "s3", "s4"]))
    )

    data = simulate(
        model,
        "dataset1",
        wanted_parameters,
        clp=clp,
        axes={"time": time, "spectral": np.array([1])},
    )

    dataset = {"dataset1": data}
    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=dataset,
        maximum_number_function_evaluations=20,
    )

    result = optimize(scheme)

    result_data = result.data["dataset1"]
    print(result_data.clp_label)
    print(result_data.clp)
    #  TODO: save reduced clp
    #  assert result_data.clp.shape == (1, 1)

    print(result_data.species_associated_spectra)
    assert result_data.species_associated_spectra.shape == (1, 2)
    assert result_data.species_associated_spectra[0, 1] == 0



from __future__ import annotations

import xarray as xr

from glotaran.analysis.problem import Problem
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.builtin.models.kinetic_image.kinetic_baseline_megacomplex import (
    KineticBaselineMegacomplex,
)
from glotaran.builtin.models.kinetic_image.kinetic_decay_megacomplex import KineticDecayMegacomplex


def finalize_kinetic_image_result(model, problem: Problem, data: dict[str, xr.Dataset]):

    for label, dataset in data.items():

        dataset_descriptor = problem.filled_dataset_descriptors[label]

        retrieve_species_associated_data(problem.model, dataset, dataset_descriptor, "images")
        retrieve_decay_associated_data(problem.model, dataset, dataset_descriptor, "images")

        if any(
            isinstance(megacomplex, KineticBaselineMegacomplex)
            for megacomplex in dataset_descriptor.megacomplex
        ):
            dataset["baseline"] = dataset.clp.sel(clp_label=f"{dataset_descriptor.label}_baseline")

        retrieve_irf(problem.model, dataset, dataset_descriptor, "images")


def retrieve_species_associated_data(model, dataset, dataset_descriptor, name):
    compartments = dataset_descriptor.initial_concentration.compartments

    dataset.coords["species"] = compartments
    dataset[f"species_associated_{name}"] = (
        (
            model.global_dimension,
            "species",
        ),
        dataset.clp.sel(clp_label=compartments).data,
    )

    if len(dataset.matrix.shape) == 3:
        #  index dependent
        dataset["species_concentration"] = (
            (
                model.global_dimension,
                model.model_dimension,
                "species",
            ),
            dataset.matrix.sel(clp_label=compartments).values,
        )
    else:
        #  index independent
        dataset["species_concentration"] = (
            (
                model.model_dimension,
                "species",
            ),
            dataset.matrix.sel(clp_label=compartments).values,
        )


def retrieve_decay_associated_data(model, dataset, dataset_descriptor, name):
    # get_das
    all_das = []
    all_a_matrix = []
    all_k_matrix = []
    all_k_matrix_reduced = []
    all_das_labels = []
    for megacomplex in dataset_descriptor.megacomplex:

        if isinstance(megacomplex, KineticDecayMegacomplex):
            k_matrix = megacomplex.full_k_matrix()

            compartments = dataset_descriptor.initial_concentration.compartments
            compartments = [c for c in compartments if c in k_matrix.involved_compartments()]

            matrix = k_matrix.full(compartments)
            matrix_reduced = k_matrix.reduced(compartments)
            a_matrix = k_matrix.a_matrix(dataset_descriptor.initial_concentration)
            rates = k_matrix.rates(dataset_descriptor.initial_concentration)
            lifetimes = 1 / rates

            das = (
                dataset[f"species_associated_{name}"].sel(species=compartments).values @ a_matrix.T
            )

            component_coords = {"rate": ("component", rates), "lifetime": ("component", lifetimes)}

            das_coords = component_coords.copy()
            das_coords[model.global_dimension] = dataset.coords[model.global_dimension]
            all_das_labels.append(megacomplex.label)
            all_das.append(
                xr.DataArray(das, dims=(model.global_dimension, "component"), coords=das_coords)
            )
            a_matrix_coords = component_coords.copy()
            a_matrix_coords["species"] = compartments
            all_a_matrix.append(
                xr.DataArray(a_matrix, coords=a_matrix_coords, dims=("component", "species"))
            )
            all_k_matrix.append(
                xr.DataArray(
                    matrix, coords=[("to_species", compartments), ("from_species", compartments)]
                )
            )

            all_k_matrix_reduced.append(
                xr.DataArray(
                    matrix_reduced,
                    coords=[("to_species", compartments), ("from_species", compartments)],
                )
            )

    if all_das:
        if len(all_das) == 1:
            dataset[f"decay_associated_{name}"] = all_das[0]
            dataset["a_matrix"] = all_a_matrix[0]
            dataset["k_matrix"] = all_k_matrix[0]
            dataset["k_matrix_reduced"] = all_k_matrix_reduced[0]

        else:
            for i, das_label in enumerate(all_das_labels):
                dataset[f"decay_associated_{name}_{das_label}"] = all_das[i].rename(
                    component=f"component_{das_label}"
                )
                dataset[f"a_matrix_{das_label}"] = all_a_matrix[i].rename(
                    component=f"component_{das_label}"
                )
                dataset[f"k_matrix_{das_label}"] = all_k_matrix[i]
                dataset[f"k_matrix_reduced_{das_label}"] = all_k_matrix_reduced[i]


def retrieve_irf(model, dataset, dataset_descriptor, name):

    irf = dataset_descriptor.irf

    if isinstance(irf, IrfMultiGaussian):
        dataset["irf"] = (
            (model.model_dimension),
            irf.calculate(
                index=0,
                global_axis=dataset.coords[model.global_dimension],
                model_axis=dataset.coords[model.model_dimension],
            ).data,
        )



from __future__ import annotations

from typing import Dict

import numpy as np

from glotaran.model import DatasetDescriptor
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import model_attribute


@model_attribute(
    properties={
        "shape": Dict[str, str],
    }
)
class SpectralMegacomplex(Megacomplex):
    def calculate_matrix(
        self,
        model,
        dataset_descriptor: DatasetDescriptor,
        indices: dict[str, int],
        axis: dict[str, np.ndarray],
        **kwargs,
    ):

        compartments = []
        for compartment in self.shape:
            if compartment in compartments:
                raise ModelError(f"More then one shape defined for compartment '{compartment}'")
            compartments.append(compartment)

        dim1 = axis[model.model_dimension].size
        dim2 = len(self.shape)
        matrix = np.zeros((dim1, dim2))

        for i, shape in enumerate(self.shape.values()):
            matrix[:, i] += shape.calculate(axis[model.model_dimension])
        return compartments, matrix



"""This package contains compartment constraint items."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import List
from typing import Tuple

from glotaran.model import model_attribute
from glotaran.model import model_attribute_typed

if TYPE_CHECKING:
    from typing import Any

    import numpy as np

    from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model import (
        KineticSpectrumModel,
    )


@model_attribute(
    properties={
        "compartment": str,
        "interval": List[Tuple[float, float]],
    },
    has_type=True,
    no_label=True,
)
class OnlyConstraint:
    """A only constraint sets the calculated matrix row of a compartment to 0
    outside the given intervals."""

    def applies(self, index: Any) -> bool:
        """
        Returns true if the index is in one of the intervals.

        Parameters
        ----------
        index :

        Returns
        -------
        applies : bool

        """

        def applies(interval):
            return interval[0] <= index <= interval[1]

        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return not any([applies(i) for i in self.interval])


@model_attribute(
    properties={
        "compartment": str,
        "interval": List[Tuple[float, float]],
    },
    has_type=True,
    no_label=True,
)
class ZeroConstraint:
    """A zero constraint sets the calculated matrix row of a compartment to 0
    in the given intervals."""

    def applies(self, index: Any) -> bool:
        """
        Returns true if the indexx is in one of the intervals.

        Parameters
        ----------
        index :

        Returns
        -------
        applies : bool

        """

        def applies(interval):
            return interval[0] <= index <= interval[1]

        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return any([applies(i) for i in self.interval])


@model_attribute_typed(
    types={
        "only": OnlyConstraint,
        "zero": ZeroConstraint,
    },
    no_label=True,
)
class SpectralConstraint:
    """A compartment constraint is applied on one compartment on one or many
    intervals on the estimated axis type.

    There are three types: zero, equal and equal area. See the documentation of
    the respective classes for details.
    """

    pass


def apply_spectral_constraints(
    model: KineticSpectrumModel, clp_labels: list[str], matrix: np.ndarray, index: float
) -> tuple[list[str], np.ndarray]:
    for constraint in model.spectral_constraints:
        if isinstance(constraint, (OnlyConstraint, ZeroConstraint)) and constraint.applies(index):
            idx = [not label == constraint.compartment for label in clp_labels]
            clp_labels = [label for label in clp_labels if label != constraint.compartment]
            matrix = matrix[:, idx]
    return (clp_labels, matrix)



"""Helper functions to give deprecation warnings."""

from __future__ import annotations

import os
import sys
from functools import wraps
from importlib import import_module
from importlib.metadata import distribution
from types import ModuleType
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import TypeVar
from typing import cast
from warnings import warn

DecoratedCallable = TypeVar(
    "DecoratedCallable", bound=Callable[..., Any]
)  # decorated function or class

if TYPE_CHECKING:
    from typing import Sequence


class OverDueDeprecation(Exception):
    """Error thrown when a deprecation should have been removed.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    """


def glotaran_version() -> str:
    """Version of the distribution.

    This is basically the same as ``glotaran.__version__`` but independent from glotaran.
    This way all of the deprecation functionality can be used even in
    ``glotaran.__init__.py`` without moving the import below the definition of
    ``__version__`` or causeing a circular import issue.

    Returns
    -------
    str
        The version string.
    """
    return distribution("pyglotaran").version


def parse_version(version_str: str) -> tuple[int, int, int]:
    """Parse version string to tuple of three ints for comparison.

    Parameters
    ----------
    version_str : str
        Fully qualified version string of the form 'major.minor.patch'.

    Returns
    -------
    tuple[int, int, int]
        Version as tuple.

    Raises
    ------
    ValueError
        If ``version_str`` has less that three elements separated by ``.``.
    ValueError
        If ``version_str`` 's first three elements can not be casted to int.
    """
    error_message = (
        "version_str need to be a fully qualified version consisting of "
        f"int parts (e.g. '0.0.1'), got {version_str!r}"
    )
    split_version = version_str.partition("-")[0].split(".")
    if len(split_version) < 3:
        raise ValueError(error_message)
    try:
        return tuple(map(int, split_version[:3]))  # type:ignore [return-value]
    except ValueError:
        raise ValueError(error_message)


def check_qualnames_in_tests(qual_names: Sequence[str], importable_indices: Sequence[int]):
    """Test that qualnames import path exists when running tests.

    All deprecations should be tested anyway in order to get the proper
    errors when a deprecation is overdue.
    This helperfunction also helps to ensure that at least the import
    paths (``qual_names``) of the old and new usage exist.

    Parameters
    ----------
    qual_names : Sequence[str]
        Sequence of fully qualified module attribute names,
        optionally with call arguments.
    importable_indices: Sequence[int]
        Indices of corresponding to ``qual_names`` indicating
        how to slice each ``qual_name`` split at ``.``, for the import
        and attribute checking.

    See Also
    --------
    warn_deprecated
    deprecate
    """
    # Since this is always true for tests run with pytest we ignore the branch coverage
    if "PYTEST_CURRENT_TEST" in os.environ:  # pragma: no branch
        for qual_name, slice_index in zip(qual_names, importable_indices):
            qual_name_parts = qual_name.partition("(")[0].partition("[")[0].split(".")
            module_name = ".".join(qual_name_parts[:-slice_index])
            object_name = qual_name_parts[-slice_index]
            module = __import__(module_name, fromlist=(object_name))
            assert hasattr(module, object_name)
            if slice_index != 1:
                item = getattr(module, object_name)
                hasattr(item, qual_name_parts[-slice_index + 1])


def warn_deprecated(
    *,
    deprecated_qual_name_usage: str,
    new_qual_name_usage: str,
    to_be_removed_in_version: str,
    check_qual_names: tuple[bool, bool] = (True, True),
    stacklevel: int = 2,
    importable_indices: tuple[int, int] = (1, 1),
) -> None:
    """Raise deprecation warning with change information.

    The change information are old / new usage information and end of support version.

    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    new_qual_name_usage : str
        New usage as fully qualified name e.g.:
        ``'glotaran.io.load_model(model_yml_str, format_name="yml_str")'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    check_qual_names : tuple[bool, bool]
        Whether or not to check for the existence ``deprecated_qual_name_usage`` and
        ``deprecated_qual_name_usage``

        *   Set the first value to False to prevent infinite recursion error when changing
            a module attribute import.

        *   Set the second value to False if the new usage in in a different package or
            there is none.

    stacklevel: int
        Stack at which the warning should be shown as raise. Default: 2

    importable_indices : tuple[int, int]
        Indices from right for most nested item which is importable for
        ``deprecated_qual_name_usage`` and ``new_qual_name_usage``
        after splitting at ``.``. This is used when the old or new usage
        is a method or mapping access. E.g. let ``deprecated_qual_name_usage``
        be ``package.module.class.mapping["key"]``, then you would use
        ``importable_indices=(2, 1)``, this way func:`check_qualnames_in_tests`
        will import ``package.module.class`` and check if ``class`` has an attribute
        ``mapping``.

    Warns
    -----
    OverDueDeprecation
        If the current version is greater or equal to ``end_of_life_version``.

    See Also
    --------
    deprecate
    deprecate_module_attribute
    deprecate_submodule
    check_qualnames_in_tests

    Examples
    --------
    This is the way the old ``read_parameters_from_yaml_file`` could deprecated and the usage of
    ``load_model`` being promoted instead.


    .. code-block:: python
        :caption: glotaran/deprecation/modules/glotaran_root.py

        def read_parameters_from_yaml_file(model_path: str):
            warn_deprecated(
                deprecated_qual_name_usage="glotaran.read_parameters_from_yaml_file(model_path)",
                new_qual_name_usage="glotaran.io.load_model.load_model(model_path)",
                to_be_removed_in_version="0.6.0",
            )
            return load_model(model_path)

    """
    if (
        parse_version(glotaran_version()) >= parse_version(to_be_removed_in_version)
        and "dev" not in glotaran_version()
    ):
        raise OverDueDeprecation(
            f"Support for {deprecated_qual_name_usage.partition('(')[0]!r} was "
            f"supposed to be dropped in version: {to_be_removed_in_version!r}\n"
            f"Current version is: {glotaran_version()!r}"
        )
    qual_names = (deprecated_qual_name_usage, new_qual_name_usage)
    selected_qual_names = [
        qual_name for qual_name, check in zip(qual_names, check_qual_names) if check
    ]
    selected_indices = importable_indices[: len(selected_qual_names)]
    check_qualnames_in_tests(qual_names=selected_qual_names, importable_indices=selected_indices)
    warn(
        DeprecationWarning(
            f"Usage of {deprecated_qual_name_usage!r} was deprecated, "
            f"use {new_qual_name_usage!r} instead.\n"
            f"This usage will be an error in version: {to_be_removed_in_version!r}."
        ),
        stacklevel=stacklevel,
    )


def deprecate(
    *,
    deprecated_qual_name_usage: str,
    new_qual_name_usage: str,
    to_be_removed_in_version: str,
    has_glotaran_replacement: bool = True,
    importable_indices: tuple[int, int] = (1, 1),
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    """Decorate a function, method or class to deprecate it.

    This raises deprecation warning with old / new usage information and
    end of support version.


    Parameters
    ----------
    deprecated_qual_name_usage : str
        Old usage with fully qualified name e.g.:
        ``'glotaran.read_model_from_yaml(model_yml_str)'``
    new_qual_name_usage : str
        New usage as fully qualified name e.g.:
        ``'glotaran.io.load_model(model_yml_str, format_name="yml_str")'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.
    has_glotaran_replacement : bool
        Whether or not this functionality has a replacement in core
        pyglotaran. This will be mapped to the second entry of ``check_qualnames``
        in :func:`warn_deprecated`.
    importable_indices : Sequence[int]
        Indices from right for most nested item which is importable for
        ``deprecated_qual_name_usage`` and ``new_qual_name_usage``
        after splitting at ``.``. This is used when the old or new usage
        is a method or mapping access. E.g. let ``deprecated_qual_name_usage``
        be ``package.module.class.mapping["key"]``, then you would use
        ``importable_indices=(2, 1)``, this way func:`check_qualnames_in_tests`
        will import ``package.module.class`` and check if ``class`` has an attribute
        ``mapping``. Default


    Returns
    -------
    DecoratedCallable
        Original function or class throwing a Deprecation warning when used.

    Warns
    -----
    OverDueDeprecation
        If the current version is greater or equal to ``end_of_life_version``.

    See Also
    --------
    warn_deprecated
    deprecate_module_attribute
    deprecate_submodule
    check_qualnames_in_tests

    Examples
    --------
    This is the way the old ``read_parameters_from_yaml_file`` was deprecated and the usage of
    ``load_model`` was promoted instead.


    .. code-block:: python
        :caption: glotaran/deprecation/modules/glotaran_root.py

        @deprecate(
            deprecated_qualname_usage="glotaran.read_parameters_from_yaml_file(model_path)",
            new_qualname_usage="glotaran.io.load_model(model_path)",
            to_be_removed_in_version="0.6.0",
        )
        def read_parameters_from_yaml_file(model_path: str):
            return load_model(model_path)


    .. # noqa: DAR402
    """

    def inject_warn_into_call(deprecated_object: DecoratedCallable) -> DecoratedCallable:
        """Wrap warning into function call.

        Used on deprecated_object.__new__ if it's a class or else on deprecated_object.
        """

        @wraps(deprecated_object)
        def inner_wrapper(*args: Any, **kwargs: Any) -> DecoratedCallable:
            """Wrap running the function and warning."""
            warn_deprecated(
                deprecated_qual_name_usage=deprecated_qual_name_usage,
                new_qual_name_usage=new_qual_name_usage,
                to_be_removed_in_version=to_be_removed_in_version,
                stacklevel=3,
                check_qual_names=(True, has_glotaran_replacement),
                importable_indices=importable_indices,
            )
            return deprecated_object(*args, **kwargs)

        return cast(DecoratedCallable, inner_wrapper)

    def outer_wrapper(deprecated_object: DecoratedCallable) -> DecoratedCallable:
        """Wrap deprecated_object of all callable kinds."""
        if type(deprecated_object) is not type:
            return cast(DecoratedCallable, inject_warn_into_call(deprecated_object))

        setattr(
            deprecated_object,
            "__new__",
            inject_warn_into_call(deprecated_object.__new__),  # type: ignore [arg-type]
        )
        return deprecated_object

    return outer_wrapper


def module_attribute(module_qual_name: str, attribute_name: str) -> Any:
    """Import and return the attribute (e.g. function or class) of a module.

    This is basically the same as ``from module_name import attribute_name as return_value``
    where this function returns ``return_value``.

    Parameters
    ----------
    module_qual_name : str
        Fully qualified name for a module e.g. ``glotaran.model.base_model``
    attribute_name : str
        Name of the attribute e.g. ``Model``

    Returns
    -------
    Any
        Attribute of the module, e.g. a function or class.
    """
    module = import_module(module_qual_name)
    return getattr(module, attribute_name)


def deprecate_module_attribute(
    *,
    deprecated_qual_name: str,
    new_qual_name: str,
    to_be_removed_in_version: str,
) -> Any:
    """Import and return and anttribute from the new location.

    This needs to be wrapped in the definition of a module wide
    ``__getattr__`` function so it won't throw warnings all the time
    (see example).

    Parameters
    ----------
    deprecated_qual_name : str
        Fully qualified name of the deprecated attribute e.g.:
        ``glotaran.ParameterGroup``
    new_qual_name : str
        Fully qualified name of the new attribute e.g.:
        ``glotaran.parameter.ParameterGroup``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.

    Returns
    -------
    Any
        Module attribute from its new location.

    See Also
    --------
    deprecate
    warn_deprecated
    deprecate_submodule

    Examples
    --------
    When deprecating the usage of ``ParameterGroup`` the root of ``glotaran``
    and promoting to import it from ``glotaran.parameter`` the following code
    was added to the root ``__init__.py``.

    .. code-block:: python
        :caption: glotaran/__init__.py


        def __getattr__(attribute_name: str):
            from glotaran.deprecation import deprecate_module_attribute

            if attribute_name == "ParameterGroup":
                return deprecate_module_attribute(
                    deprecated_qual_name="glotaran.ParameterGroup",
                    new_qual_name="glotaran.parameter.ParameterGroup",
                    to_be_removed_in_version="0.6.0",
                )

            raise AttributeError(f"module {__name__} has no attribute {attribute_name}")

    """
    module_name = ".".join(new_qual_name.split(".")[:-1])
    attribute_name = new_qual_name.split(".")[-1]

    warn_deprecated(
        deprecated_qual_name_usage=deprecated_qual_name,
        new_qual_name_usage=new_qual_name,
        to_be_removed_in_version=to_be_removed_in_version,
        check_qual_names=(False, True),
        stacklevel=4,
        importable_indices=(1, 1),
    )
    return module_attribute(module_name, attribute_name)


def deprecate_submodule(
    *,
    deprecated_module_name: str,
    new_module_name: str,
    to_be_removed_in_version: str,
) -> ModuleType:
    r"""Create a module at runtime which retrieves attributes from new module.

    When moving a module, create a variable with the modules name in the
    parent packages ``__init__.py``, so imports will be redirected to the
    new module location and a deprecation warning will be given, to help
    the user adjust the outdated code.
    Each time an attribute is retrieved there will be a deprecation warning.

    Parameters
    ----------
    deprecated_module_name : str
        Fully qualified name of the deprecated module e.g.:
        ``'glotaran.analysis.result'``
    new_module_name : str
        Fully qualified name of the new module e.g.:
        ``'glotaran.project.result'``
    to_be_removed_in_version : str
        Version the support for this usage will be removed.

    Returns
    -------
    ModuleType
        Module containing

    See Also
    --------
    deprecate
    deprecate_module_attribute

    Examples
    --------
    When moving the module ``result`` from ``glotaran.analysis.result`` to
    ``glotaran.project.result`` the following code was added to the old parent
    packages (``glotaran.analysis``) ``__init__.py``.

    .. code-block:: python
        :caption: glotaran/analysis/__init__.py

        from glotaran.deprecation.deprecation_utils import deprecate_submodule

        result = deprecate_submodule(
            deprecated_module_name="glotaran.analysis.result",
            new_module_name="glotaran.project.result",
            to_be_removed_in_version="0.6.0",
        )
    """
    new_module = import_module(new_module_name)
    deprecated_module = ModuleType(
        deprecated_module_name,
        f"Deprecated use {new_module_name!r} instead.\n\n{new_module.__doc__}",
    )

    def warn_getattr(attribute_name: str):

        if attribute_name == "__file__":
            return module_attribute(new_module, attribute_name)

        elif attribute_name in dir(new_module):
            return deprecate_module_attribute(
                deprecated_qual_name=f"{deprecated_module_name}.{attribute_name}",
                new_qual_name=f"{new_module_name}.{attribute_name}",
                to_be_removed_in_version=to_be_removed_in_version,
            )

        raise AttributeError(f"module {deprecated_module_name} has no attribute {attribute_name}")

    setattr(deprecated_module, "__getattr__", warn_getattr)
    setattr(deprecated_module, "__package__", deprecated_module_name.split(".")[:-1])
    setattr(deprecated_module, "__dir__", new_module.__dir__)

    sys.modules[deprecated_module_name] = deprecated_module
    return deprecated_module



from __future__ import annotations

from typing import Any

import numpy as np

from glotaran.builtin.models.kinetic_spectrum.kinetic_spectrum_model import KineticSpectrumModel
from glotaran.model import model_attribute
from glotaran.model import model_attribute_typed

class OnlyConstraint:
    @property
    def compartment(self) -> str: ...
    @property
    def interval(self) -> list[tuple[float, float]]: ...
    def applies(self, index: Any) -> bool: ...

class ZeroConstraint:
    @property
    def compartment(self) -> str: ...
    @property
    def interval(self) -> list[tuple[float, float]]: ...
    def applies(self, index: Any) -> bool: ...

class SpectralConstraint: ...  # noqa: E701

def apply_spectral_constraints(
    model: KineticSpectrumModel, clp_labels: list[str], matrix: np.ndarray, index: float
) -> tuple[list[str], np.ndarray]: ...



from __future__ import annotations

from glotaran.model import DatasetDescriptor
from glotaran.model import model_attribute

class KineticImageDatasetDescriptor(DatasetDescriptor):
    @property
    def initial_concentration(self) -> str: ...
    @property
    def irf(self) -> str: ...
    @property
    def baseline(self) -> bool: ...
    def get_k_matrices(self): ...
    def compartments(self): ...



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING
from typing import Deque
from typing import Dict
from typing import NamedTuple
from typing import TypeVar

import numpy as np
import xarray as xr

from glotaran.analysis.nnls import residual_nnls
from glotaran.analysis.util import get_min_max_from_interval
from glotaran.analysis.variable_projection import residual_variable_projection
from glotaran.io.prepare_dataset import add_svd_to_dataset
from glotaran.model import DatasetDescriptor
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme

if TYPE_CHECKING:
    from typing import Hashable


class ParameterError(ValueError):
    def __init__(self):
        super().__init__("Parameter not initialized")


class UngroupedProblemDescriptor(NamedTuple):
    dataset: DatasetDescriptor
    data: xr.DataArray
    model_axis: np.ndarray
    global_axis: np.ndarray
    weight: xr.DataArray


class GroupedProblemDescriptor(NamedTuple):
    label: str
    indices: dict[str, int]
    axis: dict[str, np.ndarray]


class ProblemGroup(NamedTuple):
    data: np.ndarray
    weight: np.ndarray
    has_scaling: bool
    """Indicates if at least one dataset in the group needs scaling."""
    group: str
    """The concatenated labels of the involved datasets."""
    data_sizes: list[int]
    """Holds the sizes of the concatenated datasets."""
    descriptor: list[GroupedProblemDescriptor]


UngroupedBag = Dict[str, UngroupedProblemDescriptor]
GroupedBag = Deque[ProblemGroup]

XrDataContainer = TypeVar("XrDataContainer", xr.DataArray, xr.Dataset)


class Problem:
    """A Problem class"""

    def __init__(self, scheme: Scheme):
        """Initializes the Problem class from a scheme (:class:`glotaran.analysis.scheme.Scheme`)

        Args:
            scheme (Scheme): An instance of :class:`glotaran.analysis.scheme.Scheme`
                which defines your model, parameters, and data
        """

        self._scheme = scheme

        self._model = scheme.model
        self._global_dimension = scheme.model.global_dimension
        self._model_dimension = scheme.model.model_dimension
        self._prepare_data(scheme.data)

        self._index_dependent = scheme.model.index_dependent()
        self._grouped = scheme.model.grouped()
        self._bag = None
        self._groups = None

        self._residual_function = (
            residual_nnls if scheme.non_negative_least_squares else residual_variable_projection
        )
        self._parameters = None
        self._filled_dataset_descriptors = None

        self.parameters = scheme.parameters.copy()
        self._parameter_history = []

        # all of the above are always not None

        self._clp_labels = None
        self._matrices = None
        self._reduced_clp_labels = None
        self._reduced_matrices = None
        self._reduced_clps = None
        self._clps = None
        self._weighted_residuals = None
        self._residuals = None
        self._additional_penalty = None
        self._full_axis = None
        self._full_penalty = None

    @property
    def scheme(self) -> Scheme:
        """Property providing access to the used scheme

        Returns:
            Scheme: An instance of :class:`glotaran.analysis.scheme.Scheme`
                Provides access to data, model, parameters and optimization arguments.
        """
        return self._scheme

    @property
    def model(self) -> Model:
        """Property providing access to the used model

        The model is a subclass of :class:`glotaran.model.Model` decorated with the `@model`
        decorator :class:`glotaran.model.model_decorator.model`
        For an example implementation see e.g. :class:`glotaran.builtin.models.kinetic_spectrum`

        Returns:
            Model: A subclass of :class:`glotaran.model.Model`
                The model must be decorated with the `@model` decorator
                :class:`glotaran.model.model_decorator.model`
        """
        return self._model

    @property
    def data(self) -> dict[str, xr.Dataset]:
        return self._data

    @property
    def parameters(self) -> ParameterGroup:
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: ParameterGroup):
        self._parameters = parameters
        self.reset()

    @property
    def parameter_history(self) -> list[ParameterGroup]:
        return self._parameter_history

    @property
    def grouped(self) -> bool:
        return self._grouped

    @property
    def index_dependent(self) -> bool:
        return self._index_dependent

    @property
    def filled_dataset_descriptors(self) -> dict[str, DatasetDescriptor]:
        return self._filled_dataset_descriptors

    @property
    def bag(self) -> UngroupedBag | GroupedBag:
        if not self._bag:
            self.init_bag()
        return self._bag

    @property
    def groups(self) -> dict[str, list[str]]:
        if not self._groups and self._grouped:
            self.init_bag()
        return self._groups

    @property
    def clp_labels(
        self,
    ) -> dict[str, list[str] | list[list[str]]]:
        if self._clp_labels is None:
            self.calculate_matrices()
        return self._clp_labels

    @property
    def matrices(
        self,
    ) -> dict[str, np.ndarray | list[np.ndarray]]:
        if self._matrices is None:
            self.calculate_matrices()
        return self._matrices

    @property
    def reduced_clp_labels(
        self,
    ) -> dict[str, list[str] | list[list[str]]]:
        if self._reduced_clp_labels is None:
            self.calculate_matrices()
        return self._reduced_clp_labels

    @property
    def reduced_matrices(
        self,
    ) -> dict[str, np.ndarray] | dict[str, list[np.ndarray]] | list[np.ndarray]:
        if self._reduced_matrices is None:
            self.calculate_matrices()
        return self._reduced_matrices

    @property
    def reduced_clps(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._reduced_clps is None:
            self.calculate_residual()
        return self._reduced_clps

    @property
    def clps(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._clps is None:
            self.calculate_residual()
        return self._clps

    @property
    def weighted_residuals(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._weighted_residuals is None:
            self.calculate_residual()
        return self._weighted_residuals

    @property
    def residuals(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._residuals is None:
            self.calculate_residual()
        return self._residuals

    @property
    def additional_penalty(
        self,
    ) -> dict[str, list[float]]:
        if self._additional_penalty is None:
            self.calculate_additional_penalty()
        return self._additional_penalty

    @property
    def full_penalty(self) -> np.ndarray:
        if self._full_penalty is None:
            residuals = self.weighted_residuals
            additional_penalty = self.additional_penalty
            if not self.grouped:
                residuals = [np.concatenate(residuals[label]) for label in residuals.keys()]

            self._full_penalty = (
                np.concatenate((np.concatenate(residuals), additional_penalty))
                if additional_penalty is not None
                else np.concatenate(residuals)
            )
        return self._full_penalty

    @property
    def cost(self) -> float:
        return np.sum(self._full_penalty)

    def save_parameters_for_history(self):
        self._parameter_history.append(self._parameters)

    def reset(self):
        """Resets all results and `DatasetDescriptors`. Use after updating parameters."""
        self._filled_dataset_descriptors = {
            label: descriptor.fill(self._model, self._parameters)
            for label, descriptor in self._model.dataset.items()
        }
        self._reset_results()

    def _reset_results(self):
        self._clp_labels = None
        self._matrices = None
        self._reduced_clp_labels = None
        self._reduced_matrices = None
        self._reduced_clps = None
        self._clps = None
        self._weighted_residuals = None
        self._residuals = None
        self._additional_penalty = None
        self._full_penalty = None

    def _prepare_data(self, data: dict[str, xr.DataArray | xr.Dataset]):
        self._data = {}
        for label, dataset in data.items():
            if self._model_dimension not in dataset.dims:
                raise ValueError(
                    "Missing coordinates for dimension "
                    f"'{self._model_dimension}' in data for dataset "
                    f"'{label}'"
                )
            if self._global_dimension not in dataset.dims:
                raise ValueError(
                    "Missing coordinates for dimension "
                    f"'{self._global_dimension}' in data for dataset "
                    f"'{label}'"
                )
            if isinstance(dataset, xr.DataArray):
                dataset = dataset.to_dataset(name="data")

            lsv_dim, rsv_dim = self._preferred_svd_dims(dataset)

            if lsv_dim == "time" and rsv_dim == "spectral":
                add_svd_to_dataset(dataset)

            dataset = self._transpose_dataset(
                dataset, ordered_dims=[self._model_dimension, self._global_dimension]
            )

            if lsv_dim != "time" or rsv_dim != "spectral":
                add_svd_to_dataset(dataset, lsv_dim=lsv_dim, rsv_dim=rsv_dim)

            self._add_weight(label, dataset)
            self._data[label] = dataset

    def _preferred_svd_dims(self, datacontainer: XrDataContainer) -> tuple[str, str]:
        """Get preferrer dimension for the SVD, with fallback to model and global dimensions.

        Parameters
        ----------
        dataset : XrDataContainer
            Dataset to check if the preferred dimensions exist in.

        Returns
        -------
        tuple[str, str]: (lsv_dim, rsv_dim)
            lsv_dim: Dimension of the left singular vectors.
            rsv_dim: Dimension of the right singular vectors.
        """

        lsv_dim = "time" if "time" in datacontainer.dims else self._model_dimension
        rsv_dim = "spectral" if "spectral" in datacontainer.dims else self._global_dimension

        return lsv_dim, rsv_dim

    def _transpose_dataset(
        self, datacontainer: XrDataContainer, ordered_dims: list[Hashable]
    ) -> XrDataContainer:
        """Reorder dimension of the datacontainer with the order provided by ``ordered_dims``.

        Parameters
        ----------
        dataset: XrDataContainer
            Dataset to be reordered
        ordered_dims: list[Hashable]
            Order the dimensions should be in.

        Returns
        -------
        XrDataContainer
            Datacontainer with reordered dimensions.
        """
        ordered_dims = list(filter(lambda dim: dim in datacontainer.dims, ordered_dims))
        ordered_dims += list(filter(lambda dim: dim not in ordered_dims, datacontainer.dims))
        return datacontainer.transpose(*ordered_dims)

    def _add_weight(self, label, dataset):

        # if the user supplies a weight we ignore modeled weights
        if "weight" in dataset:
            if any(label in weight.datasets for weight in self.model.weights):
                warnings.warn(
                    f"Ignoring model weight for dataset '{label}'"
                    " because weight is already supplied by dataset."
                )
            return

        global_axis = dataset.coords[self.model.global_dimension]
        model_axis = dataset.coords[self.model.model_dimension]

        for weight in self.model.weights:
            if label in weight.datasets:
                if "weight" not in dataset:
                    dataset["weight"] = xr.DataArray(
                        np.ones_like(dataset.data), coords=dataset.data.coords
                    )

                idx = {}
                if weight.global_interval is not None:
                    idx[self.model.global_dimension] = get_min_max_from_interval(
                        weight.global_interval, global_axis
                    )
                if weight.model_interval is not None:
                    idx[self.model.model_dimension] = get_min_max_from_interval(
                        weight.model_interval, model_axis
                    )
                dataset.weight[idx] *= weight.value

    def calculate_matrices(self):
        if self._parameters is None:
            raise ParameterError
        if self.index_dependent:
            self.calculate_index_dependent_matrices()
        else:
            self.calculate_index_independent_matrices()

    def calculate_residual(self):
        if self._index_dependent:
            self.calculate_index_dependent_residual()
        else:
            self.calculate_index_independent_residual()

    def calculate_additional_penalty(self) -> np.ndarray | dict[str, np.ndarray]:
        """Calculates additional penalties by calling the model.additional_penalty function."""
        if (
            callable(self.model.has_additional_penalty_function)
            and self.model.has_additional_penalty_function()
        ):
            self._additional_penalty = self.model.additional_penalty_function(
                self.parameters,
                self.clp_labels,
                self.clps,
                self.matrices,
                self.data,
                self._scheme.group_tolerance,
            )
        else:
            self._additional_penalty = None
        return self._additional_penalty

    def create_result_data(
        self, copy: bool = True, history_index: int | None = None
    ) -> dict[str, xr.Dataset]:

        if history_index is not None and history_index != -1:
            self.parameters = self.parameter_history[history_index]
        result_data = {label: self.create_result_dataset(label, copy=copy) for label in self.data}

        if callable(self.model.finalize_data):
            self.model.finalize_data(self, result_data)

        return result_data

    def create_result_dataset(self, label: str, copy: bool = True) -> xr.Dataset:
        dataset = self.data[label]
        if copy:
            dataset = dataset.copy()
        if self.index_dependent:
            dataset = self.create_index_dependent_result_dataset(label, dataset)
        else:
            dataset = self.create_index_independent_result_dataset(label, dataset)

        self._create_svd("weighted_residual", dataset)
        self._create_svd("residual", dataset)

        # Calculate RMS
        size = dataset.residual.shape[0] * dataset.residual.shape[1]
        dataset.attrs["root_mean_square_error"] = np.sqrt(
            (dataset.residual ** 2).sum() / size
        ).values
        size = dataset.weighted_residual.shape[0] * dataset.weighted_residual.shape[1]
        dataset.attrs["weighted_root_mean_square_error"] = np.sqrt(
            (dataset.weighted_residual ** 2).sum() / size
        ).values

        # reconstruct fitted data
        dataset["fitted_data"] = dataset.data - dataset.residual
        return dataset

    def _create_svd(self, name: str, dataset: xr.Dataset):
        """Calculate the SVD of a data matrix in the dataset and add it to the dataset.

        Parameters
        ----------
        name : str
            Name of the data matrix.
        dataset : xr.Dataset
            Dataset containing the data, which will be updated with the SVD values.
        """
        lsv_dim, rsv_dim = self._preferred_svd_dims(dataset[name])
        data_array: xr.DataArray = self._transpose_dataset(
            dataset[name],
            ordered_dims=[lsv_dim, rsv_dim],
        )

        add_svd_to_dataset(
            dataset, name=name, lsv_dim=lsv_dim, rsv_dim=rsv_dim, data_array=data_array
        )

    def init_bag(self):
        """Initializes a problem bag."""
        raise NotImplementedError

    def calculate_index_dependent_matrices(
        self,
    ) -> tuple[
        dict[str, list[list[str]]],
        dict[str, list[np.ndarray]],
        dict[str, list[str]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the index dependent model matrices."""
        raise NotImplementedError

    def calculate_index_independent_matrices(
        self,
    ) -> tuple[
        dict[str, list[str]],
        dict[str, np.ndarray],
        dict[str, list[str]],
        dict[str, np.ndarray],
    ]:
        """Calculates the index independent model matrices."""
        raise NotImplementedError

    def calculate_index_dependent_residual(
        self,
    ) -> tuple[
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the index dependent residuals."""
        raise NotImplementedError

    def calculate_index_independent_residual(
        self,
    ) -> tuple[
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
        dict[str, list[np.ndarray]],
    ]:
        """Calculates the index independent residuals."""
        raise NotImplementedError

    def create_index_dependent_result_dataset(self, label: str, dataset: xr.Dataset) -> xr.Dataset:
        """Creates a result datasets for index dependent matrices."""
        raise NotImplementedError

    def create_index_independent_result_dataset(
        self, label: str, dataset: xr.Dataset
    ) -> xr.Dataset:
        """Creates a result datasets for index independent matrices."""
        raise NotImplementedError



"""Test deprecated functionality in 'glotaran.project.schmeme'."""
from __future__ import annotations

from typing import TYPE_CHECKING

import xarray as xr

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.project.scheme import Scheme

if TYPE_CHECKING:
    from pathlib import Path


def test_Scheme_from_yaml_file_method(tmp_path: Path):
    """Create Scheme from file."""
    scheme_path = tmp_path / "scheme.yml"

    model_path = tmp_path / "model.yml"
    model_path.write_text("type: kinetic-spectrum\ndataset:\n  dataset1:\n    megacomplex: []")

    parameter_path = tmp_path / "parameters.yml"
    parameter_path.write_text("[1.0, 67.0]")

    dataset_path = tmp_path / "dataset.nc"
    xr.DataArray([[1, 2, 3]], coords=[("e", [1]), ("c", [1, 2, 3])]).to_dataset(
        name="data"
    ).to_netcdf(dataset_path)

    scheme_path.write_text(
        f"""
        model: {model_path}
        parameters: {parameter_path}
        non-negative-least-squares: True
        maximum-number-function-evaluations: 42
        data:
            dataset1: {dataset_path}"""
    )

    result = deprecation_warning_on_call_test_helper(
        Scheme.from_yaml_file, args=[str(scheme_path)], raise_exception=True
    )

    assert isinstance(result, Scheme)



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Mapping
from typing import TypeVar

import numpy as np
import xarray as xr

from glotaran.analysis.optimize import optimize  # noqa: F401
from glotaran.analysis.simulation import simulate  # noqa: F401
from glotaran.model.dataset_descriptor import DatasetDescriptor
from glotaran.model.decorator import FinalizeFunction
from glotaran.model.weight import Weight
from glotaran.parameter import ParameterGroup
from glotaran.project.result import Result
from glotaran.project.scheme import Scheme  # noqa: F401

_Cls = TypeVar("_Cls")

class Model:
    _model_type: str
    dataset: Mapping[str, DatasetDescriptor]
    megacomplex: Any
    weights: Weight
    model_dimension: str
    global_dimension: str
    global_matrix = None
    finalize_data: FinalizeFunction | None = ...
    grouped: Callable[[], bool]
    index_dependent: Callable[[], bool]
    @staticmethod
    def matrix(
        dataset_descriptor: DatasetDescriptor = ..., axis=..., index=...
    ) -> tuple[None, None] | tuple[list[Any], np.ndarray]: ...
    def add_megacomplex(self, item: Any): ...
    def add_weights(self, item: Weight): ...
    def get_dataset(self, label: str) -> DatasetDescriptor: ...
    @classmethod
    def from_dict(cls: type[_Cls], model_dict_ref: dict) -> _Cls: ...
    @property
    def model_type(self) -> str: ...
    def simulate(  # noqa: F811
        self,
        dataset: str,
        parameters: ParameterGroup,
        axes: dict[str, np.ndarray] = ...,
        clp: np.ndarray | xr.DataArray = ...,
        noise: bool = ...,
        noise_std_dev: float = ...,
        noise_seed: int = ...,
    ) -> xr.Dataset: ...
    def result_from_parameter(
        self,
        parameters: ParameterGroup,
        data: dict[str, xr.DataArray | xr.Dataset],
        nnls: bool = ...,
        group_atol: float = ...,
    ) -> Result: ...
    def problem_list(self, parameters: ParameterGroup = ...) -> list[str]: ...
    def validate(self, parameters: ParameterGroup = ...) -> str: ...
    def valid(self, parameters: ParameterGroup = ...) -> bool: ...
    def markdown(
        self,
        parameters: ParameterGroup = ...,
        initial_parameters: ParameterGroup = ...,
        base_heading_level: int = ...,
    ) -> str: ...



import numpy as np
import pytest

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.io import load_model
from glotaran.io import load_parameters
from glotaran.project import Scheme

MODEL_BASE = """\
type: kinetic-spectrum
dataset:
    dataset1:
        megacomplex: [mc1]
        initial_concentration: j1
        irf: irf1
        shape:
            s1: sh1
initial_concentration:
    j1:
        compartments: [s1]
        parameters: [j.1]
megacomplex:
    mc1:
        k_matrix: [k1]
k_matrix:
    k1:
        matrix:
            (s1, s1): kinetic.1
shape:
    sh1:
        type: one
"""
MODEL_SIMPLE_IRF_DISPERSION = f"""\
{MODEL_BASE}
irf:
    irf1:
        type: spectral-gaussian
        center: irf.center
        width: irf.width
        dispersion_center: irf.dispersion_center
        center_dispersion: [irf.center_dispersion]
"""
MODEL_MULTI_IRF_DISPERSION = f"""\
{MODEL_BASE}
irf:
    irf1:
        type: spectral-multi-gaussian
        center: [irf.center]
        width: [irf.width]
        dispersion_center: irf.dispersion_center
        center_dispersion: [irf.center_dispersion1, irf.center_dispersion2]
        width_dispersion: [irf.width_dispersion]
"""

PARAMETERS_BASE = """\
j:
    - ['1', 1, {'vary': False, 'non-negative': False}]
kinetic:
    - ['1', 0.5, {'non-negative': False}]
"""

PARAMETERS_SIMPLE_IRF_DISPERSION = f"""\
{PARAMETERS_BASE}
irf:
    - ['center', 0.3]
    - ['width', 0.1]
    - ['dispersion_center', 400, {{'vary': False}}]
    - ['center_dispersion', 0.5]
"""

PARAMETERS_MULTI_IRF_DISPERSION = f"""\
{PARAMETERS_BASE}
irf:
    - ["center", 0.3]
    - ["width", 0.1]
    - ["dispersion_center", 400, {{"vary": False}}]
    - ["center_dispersion1", 0.01]
    - ["center_dispersion2", 0.001]
    - ["width_dispersion", 0.025]
"""


class SimpleIrfDispersion:
    model = load_model(MODEL_SIMPLE_IRF_DISPERSION, format_name="yml_str")
    parameters = load_parameters(PARAMETERS_SIMPLE_IRF_DISPERSION, format_name="yml_str")
    time_p1 = np.linspace(-1, 2, 50, endpoint=False)
    time_p2 = np.linspace(2, 5, 30, endpoint=False)
    time_p3 = np.geomspace(5, 10, num=20)
    time = np.concatenate([time_p1, time_p2, time_p3])
    spectral = np.arange(300, 500, 100)
    axis = {"time": time, "spectral": spectral}


class MultiIrfDispersion:
    model = load_model(MODEL_MULTI_IRF_DISPERSION, format_name="yml_str")
    parameters = load_parameters(PARAMETERS_MULTI_IRF_DISPERSION, format_name="yml_str")
    time = np.arange(-1, 5, 0.2)
    spectral = np.arange(300, 500, 100)
    axis = {"time": time, "spectral": spectral}


@pytest.mark.parametrize(
    "suite",
    [
        SimpleIrfDispersion,
        MultiIrfDispersion,
    ],
)
def test_spectral_irf(suite):

    model = suite.model
    print(model.validate())
    assert model.valid()

    parameters = suite.parameters
    print(model.validate(parameters))
    assert model.valid(parameters)

    dataset = simulate(model, "dataset1", parameters, suite.axis)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["spectral"].size)

    data = {"dataset1": dataset}

    scheme = Scheme(
        model=model,
        parameters=parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for label, param in result.optimized_parameters.all():
        assert np.allclose(param.value, parameters.get(label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]

    # print(resultdata)

    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["spectral"], resultdata["spectral"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, atol=1e-14)

    irf_max_at_start = resultdata.fitted_data.isel(spectral=0).argmax(axis=0)
    irf_max_at_end = resultdata.fitted_data.isel(spectral=-1).argmax(axis=0)
    print(f" irf_max_at_start: {irf_max_at_start}\n irf_max_at_end: {irf_max_at_end}")
    # These should not be equal due to dispersion:
    assert irf_max_at_start != irf_max_at_end

    assert "species_associated_spectra" in resultdata
    assert "decay_associated_spectra" in resultdata



from __future__ import annotations

import numpy as np

from glotaran.model import model_attribute
from glotaran.model import model_attribute_typed
from glotaran.parameter import Parameter

class SpectralShapeGaussian:
    @property
    def amplitude(self) -> Parameter: ...
    @property
    def location(self) -> Parameter: ...
    @property
    def width(self) -> Parameter: ...
    def calculate(self, axis: np.ndarray) -> np.ndarray: ...

class SpectralShapeOne:
    def calculate(self, axis: np.ndarray) -> np.ndarray: ...

class SpectralShapeZero:
    def calculate(self, axis: np.ndarray) -> np.ndarray: ...

class SpectralShape:
    @classmethod
    def add_type(cls, type_name: str, attribute_type: type) -> None: ...



import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.builtin.models.kinetic_image import KineticImageModel
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


def _create_gaussian_clp(labels, amplitudes, centers, widths, axis):
    return xr.DataArray(
        [
            amplitudes[i] * np.exp(-np.log(2) * np.square(2 * (axis - centers[i]) / widths[i]))
            for i, _ in enumerate(labels)
        ],
        coords=[("clp_label", labels), ("pixel", axis)],
    ).T


class OneComponentOneChannel:
    model = KineticImageModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_list(
        [101e-4, [1, {"vary": False, "non-negative": False}]]
    )
    wanted_parameters = ParameterGroup.from_list(
        [101e-3, [1, {"vary": False, "non-negative": False}]]
    )

    time = np.asarray(np.arange(0, 50, 1.5))
    axis = {"time": time, "pixel": np.asarray([0])}

    clp = xr.DataArray([[1]], coords=[("pixel", [0]), ("clp_label", ["s1"])])


class OneComponentOneChannelGaussianIrf:
    model = KineticImageModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["5"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "irf": {
                "irf1": {"type": "gaussian", "center": "2", "width": "3", "shift": ["4"]},
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_list(
        [101e-4, 0.1, 1, [0.1, {"vary": False}], [1, {"vary": False, "non-negative": False}]]
    )
    wanted_parameters = ParameterGroup.from_list(
        [
            [101e-3, {"non-negative": True}],
            [0.2, {"non-negative": True}],
            [2, {"non-negative": True}],
            [0.1, {"vary": False}],
            [1, {"vary": False, "non-negative": False}],
        ]
    )

    time = np.asarray(np.arange(-10, 50, 1.5))
    axis = {"time": time, "pixel": np.asarray([0])}
    clp = xr.DataArray([[1]], coords=[("pixel", [0]), ("clp_label", ["s1"])])


class OneComponentOneChannelMeasuredIrf:
    model = KineticImageModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1"], "parameters": ["2"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s1", "s1"): "1",
                    }
                }
            },
            "irf": {
                "irf1": {"type": "measured"},
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_list(
        [101e-4, [1, {"vary": False, "non-negative": False}]]
    )
    wanted_parameters = ParameterGroup.from_list(
        [101e-3, [1, {"vary": False, "non-negative": False}]]
    )

    time = np.asarray(np.arange(-10, 50, 1.5))
    axis = {"time": time, "pixel": np.asarray([0])}

    center = 0
    width = 5
    irf = (1 / np.sqrt(2 * np.pi)) * np.exp(
        -(time - center) * (time - center) / (2 * width * width)
    )
    model.irf["irf1"].irfdata = irf

    clp = xr.DataArray([[1]], coords=[("pixel", [0]), ("clp_label", ["s1"])])


class ThreeComponentParallel:
    model = KineticImageModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2", "s3"], "parameters": ["j.1", "j.1", "j.1"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s2", "s1"): "kinetic.1",
                        ("s3", "s2"): "kinetic.2",
                        ("s3", "s3"): "kinetic.3",
                    }
                }
            },
            "irf": {
                "irf1": {
                    "type": "multi-gaussian",
                    "center": ["irf.center"],
                    "width": ["irf.width"],
                },
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [
                ["1", 300e-3],
                ["2", 500e-4],
                ["3", 700e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
            "j": [["1", 1, {"vary": False, "non-negative": False}]],
        }
    )
    wanted_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [
                ["1", 301e-3],
                ["2", 502e-4],
                ["3", 705e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
            "j": [["1", 1, {"vary": False, "non-negative": False}]],
        }
    )

    time = np.arange(-10, 100, 1.5)
    pixel = np.arange(600, 750, 10)
    axis = {"time": time, "pixel": pixel}

    clp = _create_gaussian_clp(
        ["s1", "s2", "s3"], [7, 3, 30], [620, 670, 720], [10, 30, 50], pixel
    )


class ThreeComponentSequential:
    model = KineticImageModel.from_dict(
        {
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2", "s3"], "parameters": ["j.1", "j.0", "j.0"]},
            },
            "megacomplex": {
                "mc1": {"k_matrix": ["k1"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s2", "s1"): "kinetic.1",
                        ("s3", "s2"): "kinetic.2",
                        ("s3", "s3"): "kinetic.3",
                    }
                }
            },
            "irf": {
                "irf1": {
                    "type": "multi-gaussian",
                    "center": ["irf.center"],
                    "width": ["irf.width"],
                },
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["mc1"],
                },
            },
        }
    )

    initial_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
                {"non-negative": True},
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
            "j": [
                ["1", 1, {"vary": False, "non-negative": False}],
                ["0", 0, {"vary": False, "non-negative": False}],
            ],
        }
    )
    wanted_parameters = ParameterGroup.from_dict(
        {
            "kinetic": [
                ["1", 501e-3],
                ["2", 202e-4],
                ["3", 105e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
            "j": [
                ["1", 1, {"vary": False, "non-negative": False}],
                ["0", 0, {"vary": False, "non-negative": False}],
            ],
        }
    )

    time = np.asarray(np.arange(-10, 50, 1.0))
    pixel = np.arange(600, 750, 10)
    axis = {"time": time, "pixel": pixel}

    clp = _create_gaussian_clp(
        ["s1", "s2", "s3"], [7, 3, 30], [620, 670, 720], [10, 30, 50], pixel
    )


@pytest.mark.parametrize(
    "suite",
    [
        OneComponentOneChannel,
        OneComponentOneChannelGaussianIrf,
        #  OneComponentOneChannelMeasuredIrf,
        ThreeComponentParallel,
        ThreeComponentSequential,
    ],
)
@pytest.mark.parametrize("nnls", [True, False])
def test_kinetic_model(suite, nnls):

    model = suite.model
    print(model.validate())
    assert model.valid()

    wanted_parameters = suite.wanted_parameters
    print(model.validate(wanted_parameters))
    print(wanted_parameters)
    assert model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)

    print(model.markdown(initial_parameters))

    dataset = simulate(model, "dataset1", wanted_parameters, suite.axis, suite.clp)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["pixel"].size)

    data = {"dataset1": dataset}

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=20,
        non_negative_least_squares=nnls,
    )
    result = optimize(scheme)
    print(result.optimized_parameters)

    for label, param in result.optimized_parameters.all():
        assert np.allclose(param.value, wanted_parameters.get(label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["pixel"], resultdata["pixel"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data, rtol=1e-2)
    assert "species_associated_images" in resultdata
    assert "decay_associated_images" in resultdata

    if len(model.irf) != 0:
        assert "irf" in resultdata



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from glotaran.builtin.models.kinetic_image import KineticImageModel
from glotaran.model import Model
from glotaran.model import model
from glotaran.model.attribute import model_attribute
from glotaran.model.megacomplex import Megacomplex
from glotaran.plugin_system.base_registry import PluginOverwriteWarning
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.model_registration import get_model
from glotaran.plugin_system.model_registration import is_known_model
from glotaran.plugin_system.model_registration import known_model_names
from glotaran.plugin_system.model_registration import model_plugin_table
from glotaran.plugin_system.model_registration import register_model
from glotaran.plugin_system.model_registration import set_model_plugin

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture
def mocked_registry(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(
        __PluginRegistry,
        "model",
        {
            "foo": Model,
            "bar": KineticImageModel,
            "glotaran.builtin.models.kinetic_image.KineticImageModel": KineticImageModel,
        },
    )


@pytest.mark.usefixtures("mocked_registry")
def test_register_model():
    """Register new model."""
    register_model("base-model", Model)

    assert "base-model" in __PluginRegistry.model
    assert __PluginRegistry.model["base-model"] == Model
    assert "glotaran.model.base_model.Model" in __PluginRegistry.model
    assert __PluginRegistry.model["glotaran.model.base_model.Model"] == Model
    assert known_model_names(full_names=True) == sorted(
        [
            "foo",
            "bar",
            "glotaran.builtin.models.kinetic_image.KineticImageModel",
            "base-model",
            "glotaran.model.base_model.Model",
        ]
    )


@pytest.mark.usefixtures("mocked_registry")
def test_register_model_warning():
    """PluginOverwriteWarning raised pointing to correct file."""

    @model_attribute()
    class DummyAttr(Megacomplex):
        pass

    with pytest.warns(PluginOverwriteWarning, match="KineticImageModel.+bar.+Dummy") as record:

        @model(
            "bar",
            attributes={},
            megacomplex_types=DummyAttr,
            model_dimension="",
            global_dimension="",
        )
        class Dummy(Model):
            pass

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("mocked_registry")
def test_is_known_model():
    """Check if models are in registry"""
    assert is_known_model("foo")
    assert is_known_model("bar")
    assert not is_known_model("baz")


@pytest.mark.usefixtures("mocked_registry")
def test_get_model():
    """Get model from registry"""
    assert get_model("foo") == Model


@pytest.mark.usefixtures("mocked_registry")
def test_known_model_names():
    """Get model names from registry"""
    assert known_model_names() == sorted(["foo", "bar"])


@pytest.mark.usefixtures("mocked_registry")
def test_known_set_model_plugin():
    """Overwrite foo model"""
    assert get_model("foo") == Model
    set_model_plugin("foo", "glotaran.builtin.models.kinetic_image.KineticImageModel")
    assert get_model("foo") == KineticImageModel


@pytest.mark.usefixtures("mocked_registry")
def test_known_set_model_plugin_dot_in_model_name():
    """Raise error if model_name contains '.'"""
    with pytest.raises(
        ValueError,
        match=r"The value of 'model_name' isn't allowed to contain the character '\.' \.",
    ):
        set_model_plugin("foo.bar", "glotaran.builtin.models.kinetic_image.KineticImageModel")


@pytest.mark.usefixtures("mocked_registry")
def test_model_plugin_table():
    """Short model table."""
    expected = dedent(
        """\
        |  __Model name__  |
        |------------------|
        |      `bar`       |
        |      `foo`       |
        """
    )

    assert f"{model_plugin_table()}\n" == expected


@pytest.mark.usefixtures("mocked_registry")
def test_model_plugin_table_full():
    """Full Table with all extras."""
    expected = dedent(
        """\
        |                      __Model name__                       |                                __Plugin name__                                |
        |-----------------------------------------------------------|-------------------------------------------------------------------------------|
        |                           `bar`                           | `glotaran.builtin.models.kinetic_image.kinetic_image_model.KineticImageModel` |
        |                           `foo`                           |                       `glotaran.model.base_model.Model`                       |
        | `glotaran.builtin.models.kinetic_image.KineticImageModel` | `glotaran.builtin.models.kinetic_image.kinetic_image_model.KineticImageModel` |
        """  # noqa: E501
    )

    assert f"{model_plugin_table(plugin_names=True,full_names=True)}\n" == expected



"""Functions for simulating a global analysis model."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.analysis.util import calculate_matrix

if TYPE_CHECKING:
    from glotaran.model import Model
    from glotaran.parameter import ParameterGroup


def simulate(
    model: Model,
    dataset: str,
    parameters: ParameterGroup,
    axes: dict[str, np.ndarray] = None,
    clp: np.ndarray | xr.DataArray = None,
    noise=False,
    noise_std_dev=1.0,
    noise_seed=None,
):
    """Simulates a model.

    Parameters
    ----------
    model :
        The model to simulate.
    parameter :
        The parameters for the simulation.
    dataset :
        Label of the dataset to simulate
    axes :
        A dictionary with axes for simulation.
    clp :
        conditionally linear parameters. Will be used instead of `model.global_matrix` if given.
    noise :
        Add noise to the simulation.
    noise_std_dev :
        The standard deviation for noise simulation.
    noise_seed :
        The seed for the noise simulation.
    """

    if model.global_matrix is None and clp is None:
        raise ValueError(
            "Cannot simulate models without implementation for global matrix and no clp given."
        )

    filled_dataset = model.dataset[dataset].fill(model, parameters)

    model_dimension = axes[model.model_dimension]
    global_dimension = axes[model.global_dimension]

    result = xr.DataArray(
        data=0.0,
        coords=[
            (model.model_dimension, model_dimension),
            (model.global_dimension, global_dimension),
        ],
    )
    result = result.to_dataset(name="data")

    matrix = (
        [
            calculate_matrix(
                model,
                filled_dataset,
                {model.global_dimension: index},
                {model.model_dimension: model_dimension, model.global_dimension: global_dimension},
            )
            for index, _ in enumerate(global_dimension)
        ]
        if model.index_dependent()
        else calculate_matrix(
            model,
            filled_dataset,
            {},
            {model.model_dimension: model_dimension, model.global_dimension: global_dimension},
        )
    )
    if callable(model.constrain_matrix_function):
        matrix = (
            [
                model.constrain_matrix_function(dataset, parameters, clp, mat, global_dimension[i])
                for i, (clp, mat) in enumerate(matrix)
            ]
            if model.index_dependent()
            else model.constrain_matrix_function(dataset, parameters, matrix[0], matrix[1], None)
        )
    matrix = (
        [
            xr.DataArray(
                mat, coords=[(model.model_dimension, model_dimension), ("clp_label", clp_label)]
            )
            for clp_label, mat in matrix
        ]
        if model.index_dependent()
        else xr.DataArray(
            matrix[1], coords=[(model.model_dimension, model_dimension), ("clp_label", matrix[0])]
        )
    )

    if clp is not None:
        if clp.shape[0] != global_dimension.size:
            raise ValueError(
                f"Size of dimension 0 of clp ({clp.shape[0]}) != size of axis"
                f" '{model.global_dimension}' ({global_dimension.size})"
            )
        if isinstance(clp, xr.DataArray):
            if model.global_dimension not in clp.coords:
                raise ValueError(f"Missing coordinate '{model.global_dimension}' in clp.")
            if "clp_label" not in clp.coords:
                raise ValueError("Missing coordinate 'clp_label' in clp.")
        else:
            if "clp_label" not in axes:
                raise ValueError("Missing axis 'clp_label'")
            clp = xr.DataArray(
                clp,
                coords=[
                    (model.global_dimension, global_dimension),
                    ("clp_label", axes["clp_label"]),
                ],
            )
    else:
        clp_labels, clp = model.global_matrix(filled_dataset, global_dimension)
        clp = xr.DataArray(
            clp, coords=[(model.global_dimension, global_dimension), ("clp_label", clp_labels)]
        )
    for i in range(global_dimension.size):
        index_matrix = matrix[i] if model.index_dependent() else matrix
        result.data[:, i] = np.dot(
            index_matrix, clp[i].sel(clp_label=index_matrix.coords["clp_label"])
        )

    if noise:
        if noise_seed is not None:
            np.random.seed(noise_seed)
        result["data"] = (
            (model.model_dimension, model.global_dimension),
            np.random.normal(result.data, noise_std_dev),
        )

    return result



import numpy as np

from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import SimpleTestModel
from glotaran.parameter import ParameterGroup


def test_simulate_dataset():
    model = SimpleTestModel.from_dict(
        {
            "megacomplex": {"m1": {"grouped": False, "indexdependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                },
            },
        }
    )
    print(model.validate())
    assert model.valid()

    parameter = ParameterGroup.from_list([1, 1])
    print(model.validate(parameter))
    assert model.valid(parameter)

    est_axis = np.asarray([1, 1, 1, 1])
    cal_axis = np.asarray([2, 2, 2])

    data = simulate(model, "dataset1", parameter, {"e": est_axis, "c": cal_axis})
    assert np.array_equal(data["c"], cal_axis)
    assert np.array_equal(data["e"], est_axis)
    assert data.data.shape == (3, 4)
    assert np.array_equal(
        data.data,
        np.asarray(
            [
                [2, 4, 6],
                [4, 10, 16],
                [6, 16, 26],
                [8, 22, 36],
            ]
        ).T,
    )



import typing

import numpy as np

from glotaran.builtin.models.kinetic_image.irf import Irf
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.model import model_attribute
from glotaran.parameter import Parameter


@model_attribute(
    properties={
        "dispersion_center": {"type": Parameter, "allow_none": True},
        "center_dispersion": {"type": typing.List[Parameter], "default": []},
        "width_dispersion": {"type": typing.List[Parameter], "default": []},
        "model_dispersion_with_wavenumber": {"type": bool, "default": False},
    },
    has_type=True,
)
class IrfSpectralMultiGaussian(IrfMultiGaussian):
    """
    Represents a gaussian IRF.

    One width and one center is a single gauss.

    One center and multiple widths is a multiple gaussian.

    Multiple center and multiple widths is Double-, Triple- , etc. Gaussian.

    Parameters
    ----------

    label:
        label of the irf
    center:
        one or more center of the irf as parameter indices
    width:
        one or more widths of the gaussian as parameter index
    center_dispersion:
        polynomial coefficients for the dispersion of the
        center as list of parameter indices. None for no dispersion.
    width_dispersion:
        polynomial coefficients for the dispersion of the
        width as parameter indices. None for no dispersion.

    """

    def parameter(self, global_index: int, global_axis: np.ndarray):
        centers, widths, scale, shift, backsweep, backsweep_period = super().parameter(
            global_index, global_axis
        )

        index = global_axis[global_index] if global_index is not None else None

        if self.dispersion_center is not None:
            dist = (
                (1e3 / index - 1e3 / self.dispersion_center)
                if self.model_dispersion_with_wavenumber
                else (index - self.dispersion_center) / 100
            )

        if len(self.center_dispersion) != 0:
            if self.dispersion_center is None:
                raise Exception(self, f'No dispersion center defined for irf "{self.label}"')
            for i, disp in enumerate(self.center_dispersion):
                centers += disp * np.power(dist, i + 1)

        if len(self.width_dispersion) != 0:
            if self.dispersion_center is None:
                raise Exception(self, f'No dispersion center defined for irf "{self.label}"')
            for i, disp in enumerate(self.width_dispersion):
                widths = widths + disp * np.power(dist, i + 1)

        return centers, widths, scale, shift, backsweep, backsweep_period

    def calculate_dispersion(self, axis):
        dispersion = []
        for index, _ in enumerate(axis):
            center, _, _, _, _, _ = self.parameter(index, axis)
            dispersion.append(center)
        return np.asarray(dispersion).T


@model_attribute(
    properties={
        "center": Parameter,
        "width": Parameter,
    },
    has_type=True,
)
class IrfSpectralGaussian(IrfSpectralMultiGaussian):
    pass


Irf.add_type("spectral-multi-gaussian", IrfSpectralMultiGaussian)
Irf.add_type("spectral-gaussian", IrfSpectralGaussian)



"""This package contains the spectral shape item."""

import numpy as np

from glotaran.model import model_attribute
from glotaran.model import model_attribute_typed
from glotaran.parameter import Parameter


@model_attribute(
    properties={
        "amplitude": Parameter,
        "location": Parameter,
        "width": Parameter,
    },
    has_type=True,
)
class SpectralShapeGaussian:
    """A gaussian spectral shape"""

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Parameters
        ----------
        axis: np.ndarray
            The axis to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return self.amplitude * np.exp(
            -np.log(2) * np.square(2 * (axis - self.location) / self.width)
        )


@model_attribute(properties={}, has_type=True)
class SpectralShapeOne:
    """A gaussian spectral shape"""

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Parameters
        ----------
        axis: np.ndarray
            The axies to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return np.ones(axis.shape[0])


@model_attribute(properties={}, has_type=True)
class SpectralShapeZero:
    """A gaussian spectral shape"""

    def calculate(self, axis: np.ndarray) -> np.ndarray:
        """calculate calculates the shape.

        Only works after calling calling 'fill'.

        Parameters
        ----------
        axis: np.ndarray
            The axies to calculate the shape on.

        Returns
        -------
        shape: numpy.ndarray

        """
        return np.zeros(axis.shape[0])


@model_attribute_typed(
    types={
        "gaussian": SpectralShapeGaussian,
        "one": SpectralShapeOne,
        "zero": SpectralShapeZero,
    }
)
class SpectralShape:
    """Base class for spectral shapes"""



from __future__ import annotations

from glotaran.builtin.models.kinetic_image.initial_concentration import InitialConcentration
from glotaran.builtin.models.kinetic_image.irf import Irf
from glotaran.builtin.models.kinetic_image.irf import IrfMultiGaussian
from glotaran.builtin.models.kinetic_image.k_matrix import KMatrix
from glotaran.builtin.models.kinetic_image.kinetic_baseline_megacomplex import (
    KineticBaselineMegacomplex,
)
from glotaran.builtin.models.kinetic_image.kinetic_decay_megacomplex import KineticDecayMegacomplex
from glotaran.builtin.models.kinetic_image.kinetic_image_dataset_descriptor import (
    KineticImageDatasetDescriptor,
)
from glotaran.builtin.models.kinetic_image.kinetic_image_result import (
    finalize_kinetic_image_result,
)
from glotaran.model import Model
from glotaran.model import model


def index_dependent(model: KineticImageModel) -> bool:
    return any(
        isinstance(irf, IrfMultiGaussian) and irf.shift is not None for irf in model.irf.values()
    )


@model(
    "kinetic-image",
    attributes={
        "initial_concentration": InitialConcentration,
        "k_matrix": KMatrix,
        "irf": Irf,
    },
    dataset_type=KineticImageDatasetDescriptor,
    default_megacomplex_type="kinetic-decay",
    megacomplex_types={
        "kinetic-decay": KineticDecayMegacomplex,
        "kinetic-baseline": KineticBaselineMegacomplex,
    },
    model_dimension="time",
    global_dimension="pixel",
    grouped=False,
    index_dependent=index_dependent,
    finalize_data_function=finalize_kinetic_image_result,
)
class KineticImageModel(Model):
    pass



"""This package contains the kinetic megacomplex item."""
from __future__ import annotations

import numpy as np

from glotaran.model import DatasetDescriptor
from glotaran.model import Megacomplex
from glotaran.model import model_attribute


@model_attribute()
class KineticBaselineMegacomplex(Megacomplex):
    def calculate_matrix(
        self,
        model,
        dataset_descriptor: DatasetDescriptor,
        indices: dict[str, int],
        axis: dict[str, np.ndarray],
        **kwargs,
    ):
        size = axis[model.model_dimension].size
        compartments = [f"{dataset_descriptor.label}_baseline"]
        matrix = np.ones((size, 1), dtype=np.float64)
        return (compartments, matrix)



import typing

from glotaran.builtin.models.kinetic_image.kinetic_image_dataset_descriptor import (
    KineticImageDatasetDescriptor,
)
from glotaran.model import model_attribute


@model_attribute(
    properties={
        "shape": {"type": typing.Dict[str, str], "allow_none": True},
    }
)
class KineticSpectrumDatasetDescriptor(KineticImageDatasetDescriptor):
    pass
